"""
This script is used to generate a frame to send to a node.

Usage:
python frames.py [destination]

If no destination is provided, the frame will be sent to the broadcast address.

Example usage sending to broadcast address via mosquitto_pub:

while [ 1 ]; do  python examples/frames.py | xxd -r -p | base64 | mosquitto_pub -h localhost -p 1883 -t /mari/00A0/to_edge -l; done
"""

from marilib.mari_protocol import Frame, Header, MARI_BROADCAST_ADDRESS, MetricsProbePayload
from marilib.model import EdgeEvent
from rich import print
import sys

destination = sys.argv[1] if len(sys.argv) > 1 else MARI_BROADCAST_ADDRESS

header = Header(destination=destination)
frame = Frame(header=header, payload=b"NORMAL_APP_DATA")
print(frame)
frame_to_send = EdgeEvent.to_bytes(EdgeEvent.NODE_DATA) + frame.to_bytes()
print(frame_to_send.hex())

probe_payload = MetricsProbePayload()
print(probe_payload.packet_length, probe_payload)
print(probe_payload.to_bytes().hex())



from __future__ import annotations
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

from rich.columns import Columns
from rich.console import Console, Group
from rich.layout import Layout
from rich.live import Live
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from marilib.model import MariNode, TestState
from marilib.tui import MarilibTUI

if TYPE_CHECKING:
    from marilib.marilib_edge import MarilibEdge


class MarilibTUIEdge(MarilibTUI):
    """A Text-based User Interface for MarilibEdge."""

    def __init__(
        self,
        max_tables=3,
        re_render_max_freq=0.2,
        test_state: TestState | None = None,
    ):
        self.console = Console()
        self.live = Live(console=self.console, auto_refresh=False, transient=True)
        self.live.start()
        self.max_tables = max_tables
        self.re_render_max_freq = re_render_max_freq
        self.last_render_time = datetime.now()
        self.test_state = test_state

    def get_max_rows(self) -> int:
        """Calculate maximum rows based on terminal height."""
        terminal_height = self.console.height
        available_height = terminal_height - 10 - 2 - 2 - 1 - 2
        return max(2, available_height)

    def render(self, mari: "MarilibEdge"):
        """Render the TUI layout."""
        with mari.lock:
            if datetime.now() - self.last_render_time < timedelta(seconds=self.re_render_max_freq):
                return
            self.last_render_time = datetime.now()
            layout = Layout()
            layout.split(
                Layout(self.create_header_panel(mari), size=12),
                Layout(self.create_nodes_panel(mari)),
            )
            self.live.update(layout, refresh=True)

    def create_header_panel(self, mari: "MarilibEdge") -> Panel:
        """Create the header panel with gateway and network stats."""
        status = Text()

        # UART Status Line
        status.append("UART: ", style="bold cyan")
        status.append(
            "connected" if mari.serial_connected else "disconnected",
            style="bold green" if mari.serial_connected else "bold red",
        )
        if mari.serial_connected:
            status.append(
                f" via {mari.serial_interface.port} at {mari.serial_interface.baudrate} baud "
            )
        secs = int((datetime.now() - mari.last_received_serial_data_ts).total_seconds())
        status.append(
            f"(last: {secs}s ago)",
            style="bold green" if secs <= 1 else "bold red",
        )

        status.append("  |  ")

        # MQTT Status Line
        status.append("MQTT: ", style="bold cyan")
        if mari.uses_mqtt:
            status.append(
                "connected" if mari.mqtt_connected else "disconnected",
                style="bold green" if mari.mqtt_connected else "bold red",
            )
            if mari.mqtt_connected:
                status.append(f" to {mari.mqtt_interface.host}:{mari.mqtt_interface.port} ")
            mqtt_secs = int((datetime.now() - mari.last_received_mqtt_data_ts).total_seconds())
            status.append(
                f"(last: {mqtt_secs}s ago)",
                style="bold green" if mqtt_secs <= 1 else "bold red",
            )
        else:
            status.append("disabled", style="bold yellow")

        status.append("\n\nGateway:  ", style="bold cyan")
        status.append(f"0x{mari.gateway.info.address:016X}  |  ")
        status.append("Network ID: ", style="bold cyan")
        status.append(f"0x{mari.gateway.info.network_id:04X}  |  ")
        status.append("ASN: ", style="bold cyan")
        status.append(f"{mari.gateway.info.asn:020d}")

        status.append("\n\n")
        status.append("Schedule: ", style="bold cyan")
        status.append(f"#{mari.gateway.info.schedule_id} {mari.gateway.info.schedule_name} |  ")
        status.append(f"{len(mari.gateway.nodes)} / {mari.gateway.info.max_nodes} nodes  |  ")
        status.append(mari.gateway.info.repr_schedule_cells_with_colors())

        # --- Latency and PDR Display ---
        avg_latency_edge = mari.gateway.stats_avg_latency_roundtrip_node_edge_ms()
        has_latency_info = avg_latency_edge > 0

        # Check if we have PDR info by looking at the gateway averages
        avg_uart_pdr_up = mari.gateway.stats_avg_pdr_uplink_uart()
        avg_uart_pdr_down = mari.gateway.stats_avg_pdr_downlink_uart()
        has_uart_pdr_info = avg_uart_pdr_up > 0 or avg_uart_pdr_down > 0

        avg_radio_pdr_down = mari.gateway.stats_avg_pdr_downlink_radio()
        avg_radio_pdr_up = mari.gateway.stats_avg_pdr_uplink_radio()
        has_radio_pdr_info = avg_radio_pdr_down > 0 or avg_radio_pdr_up > 0

        pdr_info = "  |  PDR:" if has_uart_pdr_info or has_radio_pdr_info else ""
        radio_pdr_info = (
            f"  Radio ↓ {avg_radio_pdr_down:.1%} ↑ {avg_radio_pdr_up:.1%}"
            if has_radio_pdr_info
            else ""
        )
        uart_pdr_info = (
            f"  UART ↓ {avg_uart_pdr_down:.1%} ↑ {avg_uart_pdr_up:.1%}" if has_uart_pdr_info else ""
        )

        if has_latency_info or has_uart_pdr_info or has_radio_pdr_info:
            status.append("\n\n")

        # Display Latency
        if has_latency_info:
            status.append("Latency:  ", style="bold yellow")
            status.append(f"Avg: {avg_latency_edge:.1f}ms")

        # Display PDR
        status.append(f"{pdr_info}{radio_pdr_info}{uart_pdr_info}")

        status.append("\n\nStats:    ", style="bold yellow")
        if self.test_state and self.test_state.load > 0 and self.test_state.rate > 0:
            status.append(
                "Test load: ",
            )
            status.append(f"{self.test_state.load}% of {self.test_state.rate} pps")
            status.append("  |  ")

        stats = mari.gateway.stats
        status.append(f"Frames TX: {stats.sent_count(include_test_packets=True)}  |  ")
        status.append(f"Frames RX: {stats.received_count(include_test_packets=True)} |  ")
        status.append(f"TX/s: {stats.sent_count(1, include_test_packets=True)}  |  ")
        status.append(f"RX/s: {stats.received_count(1, include_test_packets=True)}")

        return Panel(
            status,
            title=f"[bold]MariEdge running since {mari.started_ts.strftime('%Y-%m-%d %H:%M:%S')}",
            border_style="blue",
        )

    def create_nodes_table(self, nodes: list[MariNode], title="") -> Table:
        table = Table(
            show_header=True,
            header_style="bold cyan",
            border_style="blue",
            padding=(0, 1),
            title=title,
        )
        table.add_column("Node Address", style="cyan")
        table.add_column("TX", justify="right")
        table.add_column("/s", justify="right")
        table.add_column("RX", justify="right")
        table.add_column("/s", justify="right")
        table.add_column("Radio ↓ PDR | RSSI", justify="center")
        table.add_column("Radio ↑ PDR | RSSI", justify="center")
        table.add_column("UART PDR ↓ | ↑", justify="center")
        table.add_column("Latency", justify="center")

        for node in nodes:
            lat_str = (
                f"{node.stats_avg_latency_roundtrip_node_edge_ms():.1f} ms"
                if node.stats_avg_latency_roundtrip_node_edge_ms() > 0
                else "..."
            )
            # PDR Downlink with color coding
            if node.stats_pdr_downlink_radio() > 0:
                if node.stats_pdr_downlink_radio() > 0.9:
                    pdr_down_str = f"[white]{node.stats_pdr_downlink_radio():>4.0%}[/white]"
                elif node.stats_pdr_downlink_radio() > 0.8:
                    pdr_down_str = f"[yellow]{node.stats_pdr_downlink_radio():>4.0%}[/yellow]"
                else:
                    pdr_down_str = f"[red]{node.stats_pdr_downlink_radio():>4.0%}[/red]"
            else:
                pdr_down_str = "..."

            # PDR Uplink with color coding
            if node.stats_pdr_uplink_radio() > 0:
                if node.stats_pdr_uplink_radio() > 0.9:
                    pdr_up_str = f"[white]{node.stats_pdr_uplink_radio():>4.0%}[/white]"
                elif node.stats_pdr_uplink_radio() > 0.8:
                    pdr_up_str = f"[yellow]{node.stats_pdr_uplink_radio():>4.0%}[/yellow]"
                else:
                    pdr_up_str = f"[red]{node.stats_pdr_uplink_radio():>4.0%}[/red]"
            else:
                pdr_up_str = "..."

            # PDR UART Up / Down with color coding
            if node.stats_pdr_downlink_uart() > 0:
                if node.stats_pdr_downlink_uart() > 0.9:
                    pdr_down_gw_edge_str = f"[white]{node.stats_pdr_downlink_uart():>4.0%}[/white]"
                elif node.stats_pdr_downlink_uart() > 0.8:
                    pdr_down_gw_edge_str = (
                        f"[yellow]{node.stats_pdr_downlink_uart():>4.0%}[/yellow]"
                    )
                else:
                    pdr_down_gw_edge_str = f"[red]{node.stats_pdr_downlink_uart():>4.0%}[/red]"
            else:
                pdr_down_gw_edge_str = "..."

            if node.stats_pdr_uplink_uart() > 0:
                if node.stats_pdr_uplink_uart() > 0.9:
                    pdr_up_gw_edge_str = f"[white]{node.stats_pdr_uplink_uart():>4.0%}[/white]"
                elif node.stats_pdr_uplink_uart() > 0.8:
                    pdr_up_gw_edge_str = f"[yellow]{node.stats_pdr_uplink_uart():>4.0%}[/yellow]"
                else:
                    pdr_up_gw_edge_str = f"[red]{node.stats_pdr_uplink_uart():>4.0%}[/red]"
            else:
                pdr_up_gw_edge_str = "..."

            rssi_node_str = (
                f"{node.stats_rssi_node_dbm():.0f}"
                if node.stats_rssi_node_dbm() is not None
                else "..."
            )
            rssi_gw_str = (
                f"{node.stats_rssi_gw_dbm():.0f}" if node.stats_rssi_gw_dbm() is not None else "..."
            )

            table.add_row(
                f"0x{node.address:016X}",
                str(node.stats.sent_count(include_test_packets=True)),
                str(node.stats.sent_count(1, include_test_packets=True)),
                str(node.stats.received_count(include_test_packets=True)),
                str(node.stats.received_count(1, include_test_packets=True)),
                f"{pdr_down_str} | {rssi_node_str} dBm",
                f"{pdr_up_str} | {rssi_gw_str} dBm",
                f"{pdr_down_gw_edge_str} | {pdr_up_gw_edge_str}",
                lat_str,
            )
        return table

    def create_nodes_panel(self, mari: "MarilibEdge") -> Panel:
        """Create the panel that contains the nodes table."""
        nodes = mari.gateway.nodes
        max_rows = self.get_max_rows()
        max_displayable_nodes = self.max_tables * max_rows
        nodes_to_display = nodes[:max_displayable_nodes]
        remaining_nodes = max(0, len(nodes) - max_displayable_nodes)
        tables = []
        current_table_nodes = []
        for i, node in enumerate(nodes_to_display):
            current_table_nodes.append(node)
            if len(current_table_nodes) == max_rows or i == len(nodes_to_display) - 1:
                title = f"Nodes {i - len(current_table_nodes) + 2}-{i + 1}"
                tables.append(self.create_nodes_table(current_table_nodes, title))
                current_table_nodes = []
                if len(tables) >= self.max_tables:
                    break
        if len(tables) > 1:
            content = Columns(tables, equal=True, expand=True)
        else:
            content = tables[0] if tables else Table()
        if remaining_nodes > 0:
            panel_content = Group(
                content,
                Text(
                    f"\n(...and {remaining_nodes} more nodes)",
                    style="bold yellow",
                ),
            )
        else:
            panel_content = content
        return Panel(
            panel_content,
            title="[bold]Connected Nodes",
            border_style="blue",
        )

    def close(self):
        """Clean up the live display."""
        self.live.stop()
        print("")



import time

import click
from marilib.mari_protocol import MARI_BROADCAST_ADDRESS, MARI_NET_ID_DEFAULT, DefaultPayload, Frame
from marilib.marilib_cloud import MarilibCloud
from marilib.model import EdgeEvent, GatewayInfo, MariNode
from marilib.communication_adapter import MQTTAdapter
from marilib.tui_cloud import MarilibTUICloud
from marilib.logger import MetricsLogger


def on_event(event: EdgeEvent, event_data: MariNode | Frame | GatewayInfo):
    """An event handler for the application."""
    pass


@click.command()
@click.option(
    "--mqtt-url",
    "-m",
    type=str,
    default="mqtt://localhost:1883",
    show_default=True,
    help="MQTT broker to use",
)
@click.option(
    "--network-id",
    "-n",
    type=lambda x: int(x, 16),
    default=MARI_NET_ID_DEFAULT,
    help=f"Network ID to use [default: 0x{MARI_NET_ID_DEFAULT:04X}]",
)
@click.option(
    "--send-periodic",
    "-s",
    type=float,
    default=0,
    show_default=True,
    help="Send periodic packet every N seconds (0 = disabled)",
)
@click.option(
    "--log-dir",
    default="logs",
    show_default=True,
    help="Directory to save metric log files.",
    type=click.Path(),
)
def main(mqtt_url: str, network_id: int, send_periodic: float, log_dir: str):
    """A basic example of using the MariLibCloud library."""

    mari = MarilibCloud(
        on_event,
        mqtt_interface=MQTTAdapter.from_url(mqtt_url, is_edge=False),
        logger=MetricsLogger(
            log_dir_base=log_dir, rotation_interval_minutes=1440, log_interval_seconds=1.0
        ),
        network_id=network_id,
        tui=MarilibTUICloud(),
        main_file=__file__,
    )

    try:
        if send_periodic > 0:
            normal_traffic_interval = send_periodic
            last_normal_send_time = 0

        while True:
            current_time = time.monotonic()

            mari.update()
            if (
                send_periodic > 0
                and current_time - last_normal_send_time >= normal_traffic_interval
            ):
                if mari.nodes:
                    mari.send_frame(MARI_BROADCAST_ADDRESS, DefaultPayload().to_bytes())
                last_normal_send_time = current_time

            mari.render_tui()
            time.sleep(1)

    except KeyboardInterrupt:
        pass
    finally:
        mari.close_tui()
        mari.logger.close()


if __name__ == "__main__":
    main()



import sys
import threading
import time

import click
from marilib.logger import MetricsLogger
from marilib.mari_protocol import MARI_BROADCAST_ADDRESS, Frame, DefaultPayload, DefaultPayloadType
from marilib.marilib_edge import MarilibEdge
from marilib.model import EdgeEvent, GatewayInfo, MariNode, TestState
from marilib.serial_uart import get_default_port
from marilib.tui_edge import MarilibTUIEdge
from marilib.communication_adapter import SerialAdapter, MQTTAdapter


class LoadTester(threading.Thread):
    def __init__(
        self,
        mari: MarilibEdge,
        test_state: TestState,
        stop_event: threading.Event,
    ):
        super().__init__(daemon=True)
        self.mari = mari
        self.test_state = test_state
        self._stop_event = stop_event
        self.has_rate = False
        self.delay = None

    def run(self):
        while not self._stop_event.is_set():
            # wait for gateway schedule to be available and try to compute rate
            if not self.has_rate:
                self.set_rate()
            if self.delay is None:
                self._stop_event.wait(0.1)  # fixed, waiting for gateway schedule to be available
                continue

            # once we have rate, send packets at that rate
            with self.mari.lock:
                nodes_exist = bool(self.mari.gateway.nodes)

            if nodes_exist:
                self.mari.send_frame(
                    MARI_BROADCAST_ADDRESS,
                    DefaultPayload(type_=DefaultPayloadType.METRICS_LOAD).with_filler_bytes(180),
                )
            self._stop_event.wait(self.delay)

    def set_rate(self):
        if self.test_state.load == 0:
            return
        max_rate = self.mari.get_max_downlink_rate()
        if max_rate == 0:
            sys.stderr.write("Error computing max rate")
            return
        self.test_state.rate = int(max_rate)
        packets_per_second = max_rate * (self.test_state.load / 100.0)
        self.delay = 1.0 / packets_per_second if packets_per_second > 0 else float("inf")
        self.has_rate = True


def on_event(event: EdgeEvent, event_data: MariNode | Frame | GatewayInfo):
    """An event handler for the application."""
    pass


@click.command()
@click.option(
    "--port",
    "-p",
    type=str,
    default=get_default_port(),
    show_default=True,
    help="Serial port to use (e.g., /dev/ttyACM0)",
)
@click.option(
    "--mqtt-host",
    "-m",
    type=str,
    default="",
    show_default=True,
    help="MQTT broker to use (default: empty, no cloud)",
)
@click.option(
    "--load",
    type=int,
    default=0,
    show_default=True,
    help="Load percentage to apply (0–100)",
)
@click.option(
    "--send-periodic",
    "-s",
    type=float,
    default=0,
    show_default=True,
    help="Send periodic packet every N seconds (0 = disabled)",
)
@click.option(
    "--log-dir",
    default="logs",
    show_default=True,
    help="Directory to save metric log files.",
    type=click.Path(),
)
def main(port: str | None, mqtt_host: str, load: int, send_periodic: float, log_dir: str):
    if not (0 <= load <= 100):
        sys.stderr.write("Error: --load must be between 0 and 100.\n")
        return

    test_state = TestState(
        load=load,
    )

    logger = MetricsLogger(log_dir_base=log_dir, rotation_interval_minutes=1440)

    mari = MarilibEdge(
        on_event,
        serial_interface=SerialAdapter(port),
        mqtt_interface=MQTTAdapter.from_url(mqtt_host, is_edge=True) if mqtt_host else None,
        logger=logger,
        main_file=__file__,
        tui=MarilibTUIEdge(test_state=test_state),
        metrics_probe_period=1.0,  # use a relatively frequent probe period to get more stats
    )

    stop_event = threading.Event()

    load_tester = LoadTester(mari, test_state, stop_event)
    if load > 0:
        load_tester.start()

    try:
        if send_periodic > 0:
            normal_traffic_interval = send_periodic
            last_normal_send_time = 0

        while not stop_event.is_set():
            current_time = time.monotonic()

            mari.update()

            mari.render_tui()

            if (
                send_periodic > 0
                and current_time - last_normal_send_time >= normal_traffic_interval
            ):
                if mari.nodes:
                    mari.send_frame(MARI_BROADCAST_ADDRESS, DefaultPayload().to_bytes())
                last_normal_send_time = current_time

            time.sleep(1)

    except KeyboardInterrupt:
        pass
    finally:
        stop_event.set()
        mari.metrics_test_disable()
        if load_tester.is_alive():
            load_tester.join()
        mari.close_tui()
        mari.logger.close()


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod

from marilib.model import MariNode


class MarilibBase(ABC):
    """Base class for Marilib applications."""

    @abstractmethod
    def update(self):
        """Recurrent bookkeeping. Don't forget to call this periodically on your main loop."""

    @abstractmethod
    def nodes(self) -> list[MariNode]:
        """Returns all nodes in the network."""

    @abstractmethod
    def add_node(self, address: int, gateway_address: int = None) -> MariNode | None:
        """Adds a node to the network."""

    @abstractmethod
    def remove_node(self, address: int) -> MariNode | None:
        """Removes a node from the network."""

    @abstractmethod
    def send_frame(self, dst: int, payload: bytes):
        """Sends a frame to the network."""

    @abstractmethod
    def render_tui(self):
        """Renders the TUI."""

    @abstractmethod
    def close_tui(self):
        """Closes the TUI."""



# this only works on raspberry pi
# to install the emulator sense_emu library:
# - sudo apt-get install python3-gi python3-gi-cairo
# - pip install sense-emu
# to install the real sense_hat library:
# - sudo apt-get install sense-hat

from sense_emu import SenseHat
import threading
import time

# ======================= Code for driving the sense hat =======================

s = SenseHat()
# s.set_rotation(90)
# ---------- colors ----------
r = [255, 0, 0]
w = [255, 255, 255]
g = [0, 128, 0]
y = [255, 255, 0]
b = [0, 0, 255]
p = [191, 0, 255]

# clear the screen to have a white background
s.clear(w)

# ---------- config ----------
nb_max = 102  # maximum number of nodes for this schedule -- will come from marilib
nb_row = 8  # number of pixels in a row
bg = w  # background of the bottom 6*6 area
nb_nodes_per_pixel = nb_max // nb_row

# ---------- shared state ----------
state_lock = threading.Lock()
node_changed = threading.Event()

nb_nodes = 0  # current nodes
prev_pixels = 0  # how many pixels were lit last time


# 6x6 font: each row is a 6-bit number (bit 5 = leftmost, bit 0 = rightmost)
font_6x6 = {
    " ": [0b000000, 0b000000, 0b000000, 0b000000, 0b000000, 0b000000],
    "0": [0b011110, 0b100001, 0b100001, 0b100001, 0b100001, 0b011110],
    "1": [0b001000, 0b011000, 0b001000, 0b001000, 0b001000, 0b111110],
    "2": [0b011110, 0b100001, 0b000010, 0b000100, 0b001000, 0b111111],
    "3": [0b011110, 0b100001, 0b000110, 0b000001, 0b100001, 0b011110],
    "4": [0b000100, 0b001100, 0b010100, 0b100100, 0b111111, 0b000100],
    "5": [0b111111, 0b100000, 0b111110, 0b000001, 0b100001, 0b011110],
    "6": [0b011110, 0b100000, 0b111110, 0b100001, 0b100001, 0b011110],
    "7": [0b111111, 0b000001, 0b000010, 0b000100, 0b001000, 0b010000],
    "8": [0b011110, 0b100001, 0b011110, 0b100001, 0b100001, 0b011110],
    "9": [0b011110, 0b100001, 0b100001, 0b111111, 0b000001, 0b011110],
}


def choose_color(multiplier: int):
    """
    Choose color of pixels depending on number of nodes  that joined
    """
    if multiplier > 8:
        return w
    if multiplier <= 3:
        return g
    if multiplier < 6:
        return y
    if multiplier <= 8:
        return r


def _set_top(x, color):
    s.set_pixel(x, 0, color)


def display_num_nodes():
    """
    update the top bar so that every (nb_nodes//nb_row) nodes that joined or left
    changes the bar by 1 pixel.
    example: if nb_max of nodes is 66 then 66//8 = 8 so every 8 nodes joining lights up one more pixel
    and every 8 nodes leaving resets 1 pixel
    """
    global prev_pixels

    with state_lock:
        current_nodes = nb_nodes
        old_pixels = prev_pixels

    new_pixels = current_nodes // nb_nodes_per_pixel
    # nodes have joined
    if new_pixels > old_pixels:
        # light pixels from old_pixels .. new_pixels-1
        for i in range(old_pixels, new_pixels):
            _set_top(i, choose_color(i + 1))

    # nodes left
    elif new_pixels < old_pixels:
        # clear pixels from new_pixels .. old_pixels-1
        for i in range(new_pixels, old_pixels):
            _set_top(i, bg)

    # update number of pixels
    with state_lock:
        prev_pixels = new_pixels


def display_num_nodes_thread():
    """
    Background thread: wait for node changes and update the top bar.
    """
    display_num_nodes()  # initial paint
    while True:
        node_changed.wait()
        node_changed.clear()
        display_num_nodes()


def number_to_columns(rows):
    """Convert 6 rows of 6-bit ints (a number) to a list of 6 columns of pixels
    (each column is a list of 6 booleans
    True=ON, False=OFF)."""
    cols = []
    for c in range(6):  # 0..5 (left->right)
        mask = 1 << (5 - c)
        col = [(rows[r] & mask) != 0 for r in range(6)]  # top->bottom
        cols.append(col)
    return cols


def text_to_columns(text):
    """Convert text to a list of columns with 1 blank column between the numbers."""
    stream = []
    for ch in text:
        rows = font_6x6.get(ch, font_6x6[" "])
        stream.extend(number_to_columns(rows))
        stream.append([False] * 6)  # spacing column
    return stream


def display_scrolling_message(text, fg, bg, speed, scrolling_times):
    """
    Scroll a 6x6 message across the Sense HAT.
    Draws ONLY on rows y=2..7 (leaves y=0 and y=1 unchanged).
    """
    text = " " + text
    cols = text_to_columns(text)
    cols.extend([[False] * 6 for _ in range(8)])  # right padding so it scrolls off

    width = 8
    for i in range(scrolling_times):
        for offset in range(len(cols)):
            for x in range(width):
                col_index = offset + x
                col = cols[col_index] if 0 <= col_index < len(cols) else [False] * 6

                # Map 6-tall column onto y = 2..7
                for y6 in range(6):
                    y = 2 + y6
                    s.set_pixel(x, y, fg if col[y6] else bg)
            time.sleep(speed)


def display_static_message(text, fg, bg):
    """
    Show a 6x6 message without scrolling.
    Draws only on rows y=2..7, rows 0 and 1 are not used
    """
    cols = text_to_columns(text)
    width = 8
    # draw 8 columns
    for x in range(width):
        src_index = x - 1  # shift right by 1 column
        if 0 <= src_index < len(cols):
            col = cols[src_index]
        else:
            col = [False] * 6  # blank when out of bounds

        for y6 in range(6):
            y = 2 + y6  # vertical position unchanged
            s.set_pixel(x, y, fg if col[y6] else bg)


def message_thread(
    static_text, scroll_text, fg_static, fg_scroll, bg, scroll_speed, scroll_repeats
):
    """
    alternates:
      1) draw static message; schedule
      2) scroll another message: network id (net_id) scroll_repeats times at scroll_speed
    Loops forever.
    """
    while True:
        display_static_message(static_text, fg_static, bg)
        time.sleep(3)
        display_scrolling_message(scroll_text, fg_scroll, bg, scroll_speed, scroll_repeats)
        time.sleep(0.2)


# ========================= functions just for testing =========================


def node_join(n: int):
    """
    Safely add n nodes and notify the drawer.
    """
    global nb_nodes
    with state_lock:
        nb_nodes = min(nb_nodes + n, nb_max)
    node_changed.set()


# just for testing
def node_leave(n: int):
    """
    Safely remove n nodes and notify the drawer.
    """
    global nb_nodes
    with state_lock:
        nb_nodes = max(nb_nodes - n, 0)
    node_changed.set()


# ================================ main function ===============================

if __name__ == "__main__":
    scroll = True
    speed = 0.1
    schedule = 5  # will come from marilib
    net_id = 1200  # will come from marilib
    scrolling_times = 3

    threading.Thread(target=display_num_nodes_thread, daemon=True).start()

    threading.Thread(
        target=message_thread,
        args=(str(schedule), str(net_id), b, p, w, speed, scrolling_times),
        daemon=True,
    ).start()

    # demo
    while True:
        time.sleep(0.8)
        node_join(24)  # +1 bar pixel
        time.sleep(0.8)
        node_leave(8)  # -1 bar pixel



import statistics
from collections import deque
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import IntEnum
import rich

from marilib.mari_protocol import Frame, MetricsProbePayload
from marilib.protocol import Packet, PacketFieldMetadata

# schedules taken from: https://github.com/DotBots/mari-evaluation/blob/main/simulations/radio-schedule.ipynb
SCHEDULES = {
    # schedule_id: {name, max_nodes, d_down, sf_duration_ms}
    1: {
        "name": "huge",
        "slots": "BBB" + ("UUSDUUUUSDUUU" * 11) + "U" * 0,
        "max_nodes": 102,
        "d_down": 22,
        "sf_duration": 256.88,
    },
    3: {
        "name": "big",
        "slots": "BBB" + ("UUSDUUUUSDUU" * 8) + "U" * 0,
        "max_nodes": 66,
        "d_down": 16,
        "sf_duration": 174.12,
    },
    4: {
        "name": "medium",
        "slots": "BBB" + ("UUSDUUUUSDUU" * 5) + "U" * 0,
        "max_nodes": 44,
        "d_down": 10,
        "sf_duration": 115.51,
    },
    6: {
        "name": "tiny",
        "slots": "BBB" + ("UUSDUUUUSDUU" * 1) + "U" * 0,
        "max_nodes": 10,
        "d_down": 2,
        "sf_duration": 29.31,
    },
}

EMPTY_SCHEDULE_DATA = {
    "name": "unknown",
    "slots": "",
    "max_nodes": 0,
    "d_down": 0,
    "sf_duration": 0,
}

MARI_TIMEOUT_NODE_IS_ALIVE = 3  # seconds
MARI_TIMEOUT_GATEWAY_IS_ALIVE = 3  # seconds

# MARI_PROBE_STATS_EPOCH_DURATION_ASN = 565 * 20 # about 10 seconds
MARI_PROBE_STATS_EPOCH_DURATION_ASN = 565 * 60  # about 30 seconds
# MARI_PROBE_STATS_EPOCH_DURATION_ASN = 565 * 2 # about 1 second

MARI_PROBE_STATS_MAX_LEN = 10


@dataclass
class TestState:
    rate: int = 0
    load: int = 0


class EdgeEvent(IntEnum):
    NODE_JOINED = 1
    NODE_LEFT = 2
    NODE_DATA = 3
    NODE_KEEP_ALIVE = 4
    GATEWAY_INFO = 5
    UNKNOWN = 255

    @classmethod
    def to_bytes(cls, event: "EdgeEvent") -> bytes:
        return event.value.to_bytes(1, "little")


@dataclass
class NodeInfoCloud(Packet):
    metadata: list[PacketFieldMetadata] = field(
        default_factory=lambda: [
            PacketFieldMetadata(name="address", length=8),
            PacketFieldMetadata(name="gateway_address", length=8),
        ],
        repr=False,
    )
    address: int = 0
    gateway_address: int = 0


@dataclass
class NodeInfoEdge(Packet):
    metadata: list[PacketFieldMetadata] = field(
        default_factory=lambda: [
            PacketFieldMetadata(name="address", length=8),
        ],
        repr=False,
    )
    address: int = 0

    def to_cloud(self, gateway_address: int) -> NodeInfoCloud:
        return NodeInfoCloud(address=self.address, gateway_address=gateway_address)


@dataclass
class NodeStatsReply(Packet):
    """Dataclass representing the statistics packet sent back by a node."""

    metadata: list[PacketFieldMetadata] = field(
        default_factory=lambda: [
            PacketFieldMetadata(name="rx_app_packets", length=4),
            PacketFieldMetadata(name="tx_app_packets", length=4),
        ]
    )
    rx_app_packets: int = 0
    tx_app_packets: int = 0


@dataclass
class FrameLogEntry:
    frame: Frame
    ts: datetime = field(default_factory=lambda: datetime.now())


@dataclass
class MetricsStats:
    latencies: deque = field(default_factory=lambda: deque(maxlen=50))
    # TODO: Add PDR stats

    def add_latency(self, rtt_seconds: float):
        self.latencies.append(rtt_seconds * 1000)

    @property
    def last_ms(self) -> float:
        return self.latencies[-1] if self.latencies else 0.0

    @property
    def avg_ms(self) -> float:
        return statistics.mean(self.latencies) if self.latencies else 0.0

    @property
    def min_ms(self) -> float:
        return min(self.latencies) if self.latencies else 0.0

    @property
    def max_ms(self) -> float:
        return max(self.latencies) if self.latencies else 0.0

    # TODO: Add PDR stats


@dataclass
class FrameStats:
    window_seconds: int = 240  # set window duration
    sent: deque[FrameLogEntry] = field(default_factory=deque)
    received: deque[FrameLogEntry] = field(default_factory=deque)
    cumulative_sent: int = 0
    cumulative_received: int = 0
    cumulative_sent_non_test: int = 0
    cumulative_received_non_test: int = 0

    def add_sent(self, frame: Frame):
        """Adds a sent frame, prunes old entries, and updates counters."""
        self.cumulative_sent += 1
        if not frame.is_test_packet:
            self.cumulative_sent_non_test += 1  # NOTE: do we need this?

        entry = FrameLogEntry(frame=frame)
        self.sent.append(entry)

        # remove old entries
        while self.sent and (entry.ts - self.sent[0].ts).total_seconds() > self.window_seconds:
            self.sent.popleft()

    def add_received(self, frame: Frame):
        """Adds a received frame and prunes old entries."""
        self.cumulative_received += 1
        if not frame.is_test_packet:
            self.cumulative_received_non_test += 1  # NOTE: do we need this?

        entry = FrameLogEntry(frame=frame)
        self.received.append(entry)

        # remove old entries
        while (
            self.received and (entry.ts - self.received[0].ts).total_seconds() > self.window_seconds
        ):
            self.received.popleft()

    def sent_count(self, window_secs: int = 0, include_test_packets: bool = True) -> int:
        if window_secs == 0:
            return self.cumulative_sent if include_test_packets else self.cumulative_sent_non_test

        now = datetime.now()
        if include_test_packets:
            return len([e for e in self.sent if now - e.ts < timedelta(seconds=window_secs)])
        else:
            return len(
                [
                    e
                    for e in self.sent
                    if now - e.ts < timedelta(seconds=window_secs) and not e.frame.is_test_packet
                ]
            )

    def received_count(self, window_secs: int = 0, include_test_packets: bool = True) -> int:
        if window_secs == 0:
            return (
                self.cumulative_received
                if include_test_packets
                else self.cumulative_received_non_test
            )

        now = datetime.now()
        if include_test_packets:
            entries = [e for e in self.received if now - e.ts < timedelta(seconds=window_secs)]
        else:
            entries = [
                e
                for e in self.received
                if now - e.ts < timedelta(seconds=window_secs) and not e.frame.is_test_packet
            ]
        return len(entries)

    def success_rate(self, window_secs: int = 0) -> float:
        s = self.sent_count(window_secs, include_test_packets=True)
        if s == 0:
            return 1.0
        r = self.received_count(window_secs, include_test_packets=True)
        return min(r / s, 1.0)

    def received_rssi_dbm(self, window_secs: int = 0) -> float:
        if not self.received:
            return 0

        if window_secs == 0:
            return int(self.received[-1].frame.stats.rssi_dbm) if self.received else 0
        n = datetime.now()
        d = [
            e.frame.stats.rssi_dbm
            for e in self.received
            if (n - e.ts < timedelta(seconds=window_secs))
        ]
        return int(sum(d) / len(d) if d else 0)


@dataclass
class MariNode:
    address: int
    gateway_address: int
    last_seen: datetime = field(default_factory=lambda: datetime.now())
    probe_stats: deque[MetricsProbePayload] = field(
        default_factory=lambda: deque(maxlen=MARI_PROBE_STATS_MAX_LEN)
    )  # NOTE: related to frequency of probe stats
    stats: FrameStats = field(default_factory=FrameStats)
    metrics_stats: MetricsStats = field(default_factory=MetricsStats)
    last_reported_rx_count: int = 0
    last_reported_tx_count: int = 0
    pdr_downlink: float = 0.0
    pdr_uplink: float = 0.0
    probe_tx_count: int = 0
    probe_rx_count: int = 0

    @property
    def is_alive(self) -> bool:
        return datetime.now() - self.last_seen < timedelta(seconds=MARI_TIMEOUT_NODE_IS_ALIVE)

    def save_probe_stats(self, probe_stats: MetricsProbePayload):
        # save the current probe stats
        self.probe_stats.append(probe_stats)

    @property
    def probe_stats_latest(self) -> MetricsProbePayload | None:
        if not self.probe_stats:
            return None
        return self.probe_stats[-1]

    @property
    def probe_stats_start_epoch(self) -> MetricsProbePayload | None:
        if len(self.probe_stats) < 2:
            return None
        return self.probe_stats[0]

    def probe_increment_tx_count(self) -> int:
        self.probe_tx_count += 1
        return self.probe_tx_count

    def probe_increment_rx_count(self) -> int:
        self.probe_rx_count += 1
        return self.probe_rx_count

    def stats_pdr_downlink_radio(self) -> float:
        if not self.probe_stats_latest:
            return 0
        return self.probe_stats_latest.pdr_downlink_radio(self.probe_stats_start_epoch)

    def stats_pdr_uplink_radio(self) -> float:
        if not self.probe_stats_latest:
            return 0
        return self.probe_stats_latest.pdr_uplink_radio(self.probe_stats_start_epoch)

    def stats_pdr_uplink_uart(self) -> float:
        if not self.probe_stats_latest:
            return 0
        return self.probe_stats_latest.pdr_uplink_uart(self.probe_stats_start_epoch)

    def stats_pdr_downlink_uart(self) -> float:
        if not self.probe_stats_latest:
            return 0
        return self.probe_stats_latest.pdr_downlink_uart(self.probe_stats_start_epoch)

    def stats_rssi_node_dbm(self) -> float:
        if not self.probe_stats_latest:
            return None
        return self.probe_stats_latest.rssi_at_node_dbm()

    def stats_rssi_gw_dbm(self) -> float:
        if not self.probe_stats_latest:
            return None
        return self.probe_stats_latest.rssi_at_gw_dbm()

    def stats_avg_latency_roundtrip_node_edge_ms(self) -> float:
        """Average latency between node and edge in milliseconds"""
        # compute average latency between node and edge, using all probe stats
        if not self.probe_stats:
            return 0
        return sum(p.latency_roundtrip_node_edge_ms() for p in self.probe_stats) / len(
            self.probe_stats
        )

    def stats_avg_latency_roundtrip_node_cloud_ms(self) -> float:
        """Average latency between node and cloud in milliseconds"""
        if not self.probe_stats:
            return 0
        return sum(p.latency_roundtrip_node_cloud_ms() for p in self.probe_stats) / len(
            self.probe_stats
        )

    def stats_latest_latency_roundtrip_node_edge_ms(self) -> float:
        """Last latency between node and edge in milliseconds"""
        # compute average latency between node and edge, using all probe stats
        if not self.probe_stats:
            return 0
        return self.probe_stats_latest.latency_roundtrip_node_edge_ms()

    def stats_latest_latency_roundtrip_node_cloud_ms(self) -> float:
        """Last latency between node and cloud in milliseconds"""
        if not self.probe_stats:
            return 0
        return self.probe_stats_latest.latency_roundtrip_node_cloud_ms()

    def register_received_frame(self, frame: Frame):
        self.stats.add_received(frame)

    def register_sent_frame(self, frame: Frame):
        self.stats.add_sent(frame)

    def as_node_info_cloud(self) -> NodeInfoCloud:
        return NodeInfoCloud(address=self.address, gateway_address=self.gateway_address)


@dataclass
class GatewayInfo(Packet):
    metadata: list[PacketFieldMetadata] = field(
        default_factory=lambda: [
            PacketFieldMetadata(name="address", length=8),
            PacketFieldMetadata(name="network_id", length=2),
            PacketFieldMetadata(name="schedule_id", length=1),
            PacketFieldMetadata(name="schedule_stats", length=4 * 8),  # 4 uint64_t values
            PacketFieldMetadata(name="asn", length=8),
            PacketFieldMetadata(name="timer", length=4),
        ]
    )
    address: int = 0
    network_id: int = 0
    schedule_id: int = 0
    schedule_stats: bytes = b""
    asn: int = 0
    timer: int = 0

    # NOTE: maybe move to a separate class, dedicated to schedule stuff
    def repr_schedule_stats(self):
        if not self.schedule_stats:
            return ""
        schedule_data = SCHEDULES.get(self.schedule_id)
        if not schedule_data:
            return ""
        all_bits = format(self.schedule_stats, f"0{4*8*8}b")
        all_bits = [all_bits[i : i + 8] for i in range(0, len(all_bits), 8)]
        all_bits.reverse()
        # print(">>>", reversed(all_bits[0].split("")))
        all_bits = [list(reversed(bits)) for bits in all_bits]
        # now just flatten the list
        all_bits = [item for sublist in all_bits for item in sublist]
        # FIXME: why do we need to skip the first byte?
        all_bits = all_bits[8:]
        # cut it down to the number of slots
        all_bits = all_bits[: len(schedule_data["slots"])]
        return "".join(all_bits)

    def repr_cell_nice(self, cell: str, is_used: int):
        is_used = bool(int(is_used))
        if cell == "B":
            return rich.text.Text("B", style=f'bold white on {"red" if is_used else "indian_red"}')
        elif cell == "S":
            return rich.text.Text(
                "S", style=f'bold white on {"purple" if is_used else "medium_purple2"}'
            )
        elif cell == "D":
            return rich.text.Text(
                "D", style=f'bold white on {"green" if is_used else "sea_green3"}'
            )
        elif cell == "U":
            return rich.text.Text(
                "U", style=f'bold white on {"yellow" if is_used else "light_yellow3"}'
            )

    def repr_schedule_cells_with_colors(self):
        schedule_data = SCHEDULES.get(self.schedule_id)
        if not schedule_data:
            return ""
        sched_stats = [
            self.repr_cell_nice(cell, is_used)
            for cell, is_used in zip(schedule_data["slots"], self.repr_schedule_stats())
        ]
        return rich.text.Text.assemble(*sched_stats)

    @property
    def schedule_name(self) -> str:
        schedule_data = SCHEDULES.get(self.schedule_id)
        return schedule_data["name"] if schedule_data else "unknown"

    @property
    def max_nodes(self) -> int:
        return SCHEDULES.get(self.schedule_id, EMPTY_SCHEDULE_DATA)["max_nodes"]

    @property
    def network_id_str(self) -> str:
        return f"{self.network_id:04X}"

    @property
    def schedule_uplink_cells(self) -> int:
        return SCHEDULES.get(self.schedule_id, EMPTY_SCHEDULE_DATA)["max_nodes"]

    @property
    def schedule_downlink_cells(self) -> int:
        return SCHEDULES.get(self.schedule_id, EMPTY_SCHEDULE_DATA)["slots"].count("D")


@dataclass
class MariGateway:
    info: GatewayInfo = field(default_factory=GatewayInfo)
    node_registry: dict[int, MariNode] = field(default_factory=dict)
    stats: FrameStats = field(default_factory=FrameStats)
    metrics_stats: MetricsStats = field(default_factory=MetricsStats)
    last_seen: datetime = field(default_factory=lambda: datetime.now())

    def __post_init__(self):
        self.last_seen = datetime.now()

    @property
    def nodes(self) -> list[MariNode]:
        return list(self.node_registry.values())

    @property
    def nodes_addresses(self) -> list[int]:
        return list(self.node_registry.keys())

    @property
    def is_alive(self) -> bool:
        return datetime.now() - self.last_seen < timedelta(seconds=MARI_TIMEOUT_GATEWAY_IS_ALIVE)

    def stats_avg_pdr_downlink_radio(self) -> float:
        if not self.nodes:
            return 0.0
        res = sum(n.stats_pdr_downlink_radio() for n in self.nodes) / len(self.nodes)
        return res if res >= 0 and res <= 1.0 else 0.0

    def stats_avg_pdr_uplink_radio(self) -> float:
        if not self.nodes:
            return 0.0
        res = sum(n.stats_pdr_uplink_radio() for n in self.nodes) / len(self.nodes)
        return res if res >= 0 and res <= 1.0 else 0.0

    def stats_avg_pdr_downlink_uart(self) -> float:
        if not self.nodes:
            return 0.0
        res = sum(n.stats_pdr_downlink_uart() for n in self.nodes) / len(self.nodes)
        return res if res >= 0 and res <= 1.0 else 0.0

    def stats_avg_pdr_uplink_uart(self) -> float:
        if not self.nodes:
            return 0.0
        res = sum(n.stats_pdr_uplink_uart() for n in self.nodes) / len(self.nodes)
        return res if res >= 0 and res <= 1.0 else 0.0

    def stats_avg_latency_roundtrip_node_edge_ms(self) -> float:
        if not self.nodes:
            return 0.0
        res = sum(n.stats_avg_latency_roundtrip_node_edge_ms() for n in self.nodes) / len(
            self.nodes
        )
        return res if res >= 0 else 0.0

    def stats_avg_latency_roundtrip_node_cloud_ms(self) -> float:
        if not self.nodes:
            return 0.0
        res = sum(n.stats_avg_latency_roundtrip_node_cloud_ms() for n in self.nodes) / len(
            self.nodes
        )
        return res if res >= 0 else 0.0

    def stats_latest_node_tx_count(self) -> int:
        """Returns sum of tx counts for all nodes"""
        if not self.nodes:
            return 0
        return sum(n.probe_stats_latest.node_tx_count for n in self.nodes if n.probe_stats_latest)

    def stats_latest_node_rx_count(self) -> int:
        """Returns sum of rx counts for all nodes"""
        if not self.nodes:
            return 0
        return sum(n.probe_stats_latest.node_rx_count for n in self.nodes if n.probe_stats_latest)

    def stats_latest_gw_tx_count(self) -> int:
        """Returns sum of tx counts for all nodes"""
        if not self.nodes:
            return 0
        return sum(n.probe_stats_latest.gw_tx_count for n in self.nodes if n.probe_stats_latest)

    def stats_latest_gw_rx_count(self) -> int:
        """Returns sum of rx counts for all nodes"""
        if not self.nodes:
            return 0
        return sum(n.probe_stats_latest.gw_rx_count for n in self.nodes if n.probe_stats_latest)

    def update(self):
        """Recurrent bookkeeping. Don't forget to call this periodically on your main loop."""
        self.node_registry = {
            addr: node for addr, node in self.node_registry.items() if node.is_alive
        }

    def set_info(self, info: GatewayInfo):
        self.info = info
        self.last_seen = datetime.now()

    def get_node(self, addr: int) -> MariNode | None:
        return self.node_registry.get(addr)

    def add_node(self, addr: int) -> MariNode:
        if node := self.get_node(addr):
            node.last_seen = datetime.now()
            return node
        node = MariNode(addr, self.info.address)
        self.node_registry[addr] = node
        return node

    def remove_node(self, addr: int) -> MariNode | None:
        return self.node_registry.pop(addr, None)

    def update_node_liveness(self, addr: int) -> MariNode:
        node = self.get_node(addr)
        if node:
            node.last_seen = datetime.now()
        else:
            node = self.add_node(addr)
        return node

    def register_received_frame(self, frame: Frame):
        if n := self.get_node(frame.header.source):
            n.register_received_frame(frame)
        self.stats.add_received(frame)

    def register_sent_frame(self, frame: Frame):
        self.stats.add_sent(frame)



# TODO: import this from like PyDotBot or similar

import dataclasses
import typing
from abc import ABC
from dataclasses import dataclass
from enum import IntEnum


class ProtocolPayloadParserException(Exception):
    """Exception raised on invalid or unsupported payload."""


class PacketType(IntEnum):
    """Types of MAC layer packet."""

    BEACON = 1
    JOIN_REQUEST = 2
    JOIN_RESPONSE = 4
    KEEP_ALIVE = 8
    DATA = 16


@dataclass
class PacketFieldMetadata:
    """Data class that describes a packet field metadata."""

    name: str = ""
    disp: str = ""
    length: int = 1
    signed: bool = False
    type_: typing.Any = int

    def __post_init__(self):
        if not self.disp:
            self.disp = self.name


@dataclass
class Packet(ABC):
    """Base class for packet classes."""

    @property
    def size(self) -> int:
        return sum(field.length for field in self.metadata)

    def from_bytes(self, bytes_):
        fields = dataclasses.fields(self)
        # base class makes metadata attribute mandatory so there's at least one
        # field defined in subclasses
        # first elements in fields has to be metadata
        if not fields or fields[0].name != "metadata":
            raise ValueError("metadata must be defined first")
        metadata = fields[0].default_factory()
        for idx, field in enumerate(fields[1:]):
            if metadata[idx].type_ is list:
                element_class = typing.get_args(field.type)[0]
                field_attribute = getattr(self, field.name)
                # subclass element is a list and previous attribute is called
                # "count" and should have already been retrieved from the byte
                # stream
                for _ in range(self.count):
                    element = element_class()
                    if len(bytes_) < element.size:
                        raise ValueError("Not enough bytes to parse")
                    field_attribute.append(element.from_bytes(bytes_))
                    bytes_ = bytes_[element.size :]
            elif metadata[idx].type_ in [bytes, bytearray]:
                # subclass element is bytes and previous attribute is called
                # "count" and should have already been retrieved from the byte
                # stream
                length = metadata[idx].length
                if hasattr(self, "count"):
                    length = self.count
                setattr(self, field.name, bytes_[0:length])
                bytes_ = bytes_[length:]
            else:
                length = metadata[idx].length
                if len(bytes_) < length:
                    raise ValueError("Not enough bytes to parse")
                setattr(
                    self,
                    field.name,
                    int.from_bytes(
                        bytes=bytes_[0:length],
                        signed=metadata[idx].signed,
                        byteorder="little",
                    ),
                )
                bytes_ = bytes_[length:]
        return self

    def to_bytes(self, byteorder="little") -> bytes:
        buffer = bytearray()
        metadata = dataclasses.fields(self)[0].default_factory()
        for idx, field in enumerate(dataclasses.fields(self)[1:]):
            value = getattr(self, field.name)
            if isinstance(value, list):
                for element in value:
                    buffer += element.to_bytes()
            elif isinstance(value, (bytes, bytearray)):
                buffer += value
            else:
                buffer += int(value).to_bytes(
                    length=metadata[idx].length,
                    byteorder=byteorder,
                    signed=metadata[idx].signed,
                )
        return buffer



import threading
import time
from typing import TYPE_CHECKING

from rich import print
from marilib.mari_protocol import Frame, DefaultPayloadType
from marilib.mari_protocol import MetricsProbePayload
from marilib.model import MariGateway, MariNode, MARI_PROBE_STATS_MAX_LEN

if TYPE_CHECKING:
    from marilib.marilib_edge import MarilibEdge


class MetricsTester:
    """A thread-based class to periodically test metrics to all nodes."""

    def __init__(self, marilib: "MarilibEdge", interval: float = 3):
        self.marilib = marilib
        self.set_interval(interval)
        self._stop_event = threading.Event()
        self._thread = threading.Thread(target=self._run, daemon=True)

    def set_interval(self, interval: float):
        if interval < 0 or interval > MARI_PROBE_STATS_MAX_LEN:
            raise ValueError(f"Interval must be >= 0 and <= {MARI_PROBE_STATS_MAX_LEN}")
        self.interval = interval

    def start(self):
        """Starts the metrics testing thread."""
        if self.interval < 0 or self.interval > MARI_PROBE_STATS_MAX_LEN:
            raise ValueError(f"Interval must be >= 0 and <= {MARI_PROBE_STATS_MAX_LEN}")
        if self.interval == 0:
            print("[yellow]Metrics tester disabled.[/]")
            return
        print(f"[yellow]Metrics tester started with interval {self.interval} seconds.[/]")
        self._thread.start()

    def stop(self):
        """Stops the metrics testing thread."""
        self._stop_event.set()
        if self._thread.is_alive():
            self._thread.join()
        print("[yellow]Metrics tester stopped.[/]")

    def _run(self):
        """The main loop for the testing thread."""
        # Initial delay to allow nodes to join
        self._stop_event.wait(self.interval)

        while not self._stop_event.is_set():
            nodes = list(self.marilib.nodes)
            if not nodes:
                self._stop_event.wait(self.interval)
                continue

            for node in nodes:
                if self._stop_event.is_set():
                    break
                self.send_metrics_request(node, "edge")
                # Spread the requests evenly over the interval
                sleep_duration = self.interval / len(nodes)
                self._stop_event.wait(sleep_duration)

    def timestamp_us(self) -> int:
        """Returns the current time in microseconds."""
        return int(time.time() * 1000 * 1000)

    def send_metrics_request(self, node: MariNode, marilib_type: str):
        """Sends a metrics request packet to a specific address."""
        payload = MetricsProbePayload()
        if marilib_type == "edge":
            payload.edge_tx_ts_us = self.timestamp_us()
            payload.edge_tx_count = node.probe_increment_tx_count()
        elif marilib_type == "cloud":
            payload.cloud_tx_ts_us = self.timestamp_us()
            payload.cloud_tx_count = node.probe_increment_tx_count()
        # print(f">>> sending metrics probe to {node.address:016x}: {payload}")
        payload = payload.to_bytes()
        # print(f"    size is {len(payload)} bytes: {payload.hex()}\n")
        self.marilib.send_frame(node.address, payload)

    def handle_response_edge(self, frame: Frame):
        """
        Processes a metrics response frame.
        This should be called when a LATENCY_DATA event is received.
        """
        node = self.marilib.gateway.get_node(frame.header.source)
        if not node:
            print(f"[red]Node not found: {frame.header.source:016x}[/]")
            return

        try:
            payload = MetricsProbePayload().from_bytes(frame.payload)
            if payload.type_ != DefaultPayloadType.METRICS_PROBE:
                print(f"[red]Expected METRICS_PROBE, got {payload.type_}[/]")
                return

        except Exception as e:
            print(f"[red]Error parsing metrics response: {e}[/]")
            return

        payload.edge_rx_ts_us = self.timestamp_us()
        payload.edge_rx_count = node.probe_increment_rx_count()

        node.save_probe_stats(payload)

        # print(f"<<< received metrics probe from {frame.header.source:016x}: {payload}")
        # print(f"    size is {len(frame.payload)} bytes: {frame.payload.hex()}\n")

        # print(f"    latency_roundtrip_node_edge_ms: {payload.latency_roundtrip_node_edge_ms()}")
        # print(f"    pdr_uplink_radio: {payload.pdr_uplink_radio(node.probe_stats_start_epoch)}")
        # print(f"    pdr_downlink_radio: {payload.pdr_downlink_radio(node.probe_stats_start_epoch)}")
        # print(f"    pdr_uplink_uart: {payload.pdr_uplink_uart(node.probe_stats_start_epoch)}")
        # print(f"    pdr_downlink_uart: {payload.pdr_downlink_uart(node.probe_stats_start_epoch)}")
        # print(f"    rssi_at_node_dbm: {payload.rssi_at_node_dbm()}")
        # print(f"    rssi_at_gw_dbm: {payload.rssi_at_gw_dbm()}")

        return payload

    def handle_response_cloud(self, frame: Frame, gateway: MariGateway, node: MariNode):
        """
        Processes a metrics response frame.
        This should be called when a LATENCY_DATA event is received.
        """
        try:
            payload = MetricsProbePayload().from_bytes(frame.payload)
            if payload.type_ != DefaultPayloadType.METRICS_PROBE:
                print(f"[red]Expected METRICS_PROBE, got {payload.type_}[/]")
                return

        except Exception as e:
            print(f"[red]Error parsing metrics response: {e}[/]")
            return

        payload.cloud_rx_ts_us = self.timestamp_us()
        payload.cloud_rx_count = node.probe_increment_rx_count()

        node.save_probe_stats(payload)

        # print(f"<<< received metrics probe from {frame.header.source:016x}: {payload}")
        # print(f"    size is {len(frame.payload)} bytes: {frame.payload.hex()}\n")

        # print(f"    latency_roundtrip_node_edge_ms: {payload.latency_roundtrip_node_edge_ms()}")
        # print(f"    pdr_uplink_radio: {payload.pdr_uplink_radio(node.probe_stats_start_epoch)}")
        # print(f"    pdr_downlink_radio: {payload.pdr_downlink_radio(node.probe_stats_start_epoch)}")
        # print(f"    pdr_uplink_uart: {payload.pdr_uplink_uart(node.probe_stats_start_epoch)}")
        # print(f"    pdr_downlink_uart: {payload.pdr_downlink_uart(node.probe_stats_start_epoch)}")
        # print(f"    rssi_at_node_dbm: {payload.rssi_at_node_dbm()}")
        # print(f"    rssi_at_gw_dbm: {payload.rssi_at_gw_dbm()}")

        return payload



from datetime import datetime, timedelta

from rich.columns import Columns
from rich.console import Console, Group
from rich.layout import Layout
from rich.live import Live
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from marilib import MarilibCloud
from marilib.model import MariGateway
from marilib.tui import MarilibTUI


class MarilibTUICloud(MarilibTUI):
    """A Text-based User Interface for MarilibCloud."""

    def __init__(
        self,
        max_tables=4,
        re_render_max_freq=0.2,
    ):
        self.console = Console()
        self.live = Live(console=self.console, auto_refresh=False, transient=True)
        self.live.start()
        self.max_tables = max_tables
        self.re_render_max_freq = re_render_max_freq
        self.last_render_time = datetime.now()

    def get_max_rows(self) -> int:
        """Calculate maximum rows based on terminal height."""
        terminal_height = self.console.height
        available_height = terminal_height - 10 - 2 - 2 - 1 - 2
        return max(2, available_height)

    def render(self, mari: MarilibCloud):
        """Render the TUI layout."""
        with mari.lock:
            if datetime.now() - self.last_render_time < timedelta(seconds=self.re_render_max_freq):
                return
            self.last_render_time = datetime.now()
            layout = Layout()
            layout.split(
                Layout(self.create_header_panel(mari), size=6),
                Layout(self.create_gateways_panel(mari)),
            )
            self.live.update(layout, refresh=True)

    def create_header_panel(self, mari: MarilibCloud) -> Panel:
        """Create the header panel with MQTT connection and network info."""
        status = Text()
        status.append("MarilibCloud is ", style="bold")
        status.append("connected", style="bold green")
        status.append(
            f" to MQTT broker {mari.mqtt_interface.host}:{mari.mqtt_interface.port} "
            f"at topic /mari/{mari.network_id_str}/to_cloud "
            f"since {mari.started_ts.strftime('%Y-%m-%d %H:%M:%S')}"
        )
        status.append("  |  ")
        secs = int((datetime.now() - mari.last_received_mqtt_data_ts).total_seconds())
        status.append(
            f"last received: {secs}s ago",
            style="bold green" if secs <= 1 else "bold red",
        )

        status.append("\n\nNetwork ID: ", style="bold cyan")
        status.append(f"0x{mari.network_id:04X}")
        status.append("  |  ")
        status.append("Gateways: ", style="bold cyan")
        status.append(f"{len(mari.gateways)}")
        status.append("  |  ")
        status.append("Nodes: ", style="bold cyan")
        status.append(f"{len(mari.nodes)}")

        return Panel(status, title="[bold]MarilibCloud Status", border_style="blue")

    def create_gateway_table(self, gateway: MariGateway) -> Table:
        """Create a table for a single gateway with 3 rows and 2 columns."""
        table = Table(
            show_header=False,
            border_style="blue",
            padding=(0, 1),
        )
        table.add_column("Field", style="bold", width=18, justify="right")
        table.add_column("Value")

        # Row 1: Gateway info
        node_count = f"{len(gateway.nodes)} / {gateway.info.schedule_uplink_cells}"
        schedule_info = f"#{gateway.info.schedule_id} {gateway.info.schedule_name}"

        # --- Latency and PDR Display ---
        avg_latency_edge = gateway.stats_avg_latency_roundtrip_node_edge_ms()
        has_latency_info = avg_latency_edge > 0

        # Check if we have PDR info by looking at the gateway averages
        avg_uart_pdr_up = gateway.stats_avg_pdr_uplink_uart()
        avg_uart_pdr_down = gateway.stats_avg_pdr_downlink_uart()
        has_uart_pdr_info = avg_uart_pdr_up > 0 or avg_uart_pdr_down > 0

        avg_radio_pdr_down = gateway.stats_avg_pdr_downlink_radio()
        avg_radio_pdr_up = gateway.stats_avg_pdr_uplink_radio()
        has_radio_pdr_info = avg_radio_pdr_down > 0 or avg_radio_pdr_up > 0

        latency_info = f"  |  Latency: {avg_latency_edge:.1f}ms" if has_latency_info else ""
        pdr_info = "  |  PDR:" if has_uart_pdr_info or has_radio_pdr_info else ""
        radio_pdr_info = (
            f"  Radio ↓ {avg_radio_pdr_down:.1%} ↑ {avg_radio_pdr_up:.1%}"
            if has_radio_pdr_info
            else ""
        )
        uart_pdr_info = (
            f"  UART ↓ {avg_uart_pdr_down:.1%} ↑ {avg_uart_pdr_up:.1%}" if has_uart_pdr_info else ""
        )

        table.add_row(
            f"[bold cyan]0x{gateway.info.address:016X}[/bold cyan]",
            f"Nodes: {node_count}  |  Schedule: {schedule_info}{latency_info}{pdr_info}{radio_pdr_info}{uart_pdr_info}",
        )

        # Row 2: Schedule usage
        schedule_repr = gateway.info.repr_schedule_cells_with_colors()
        table.add_row("[bold cyan]Live schedule[/bold cyan]", schedule_repr)

        # Row 3: Node list
        if gateway.nodes:
            node_addresses = [f"0x{node.address:016X}" for node in gateway.nodes]
            node_display = " ".join(node_addresses)
        else:
            node_display = "—"

        table.add_row("[bold cyan]Nodes[/bold cyan]", node_display)

        return table

    def create_gateways_panel(self, mari: MarilibCloud) -> Panel:
        """Create the panel that contains individual gateway tables."""
        gateways = list(mari.gateways.values())

        if not gateways:
            empty_table = Table(title="No Gateways Connected")
            return Panel(
                empty_table,
                title="[bold]Connected Gateways",
                border_style="blue",
            )

        # Create individual tables for each gateway
        gateway_tables = []
        max_displayable_gateways = self.max_tables
        gateways_to_display = gateways[:max_displayable_gateways]
        remaining_gateways = max(0, len(gateways) - max_displayable_gateways)

        for gateway in gateways_to_display:
            gateway_tables.append(self.create_gateway_table(gateway))

        # Arrange tables in columns
        if len(gateway_tables) > 1:
            content = Columns(gateway_tables, equal=True, expand=True)
        else:
            content = gateway_tables[0]

        if remaining_gateways > 0:
            panel_content = Group(
                content,
                Text(
                    f"\n...and {remaining_gateways} more gateway(s)",
                    style="bold yellow",
                ),
            )
        else:
            panel_content = content

        return Panel(
            panel_content,
            title="[bold]Connected Gateways",
            border_style="blue",
        )

    def close(self):
        """Clean up the live display."""
        self.live.stop()
        print("")



import time

from marilib.marilib_edge import MarilibEdge, EdgeEvent
from marilib.communication_adapter import MQTTAdapter, SerialAdapter
from marilib.serial_uart import get_default_port


def on_event(event, event_data):
    """An event handler for the application."""
    if event == EdgeEvent.GATEWAY_INFO:
        return
    print(".", end="", flush=True)


def main():
    mari_edge = MarilibEdge(
        on_event,
        serial_interface=SerialAdapter(get_default_port()),
        mqtt_interface=MQTTAdapter("localhost", 1883, is_edge=True),
    )

    while True:
        mari_edge.update()
        if not mari_edge.uses_mqtt:
            # only generate frames if not using MQTT
            for node in mari_edge.nodes:
                mari_edge.send_frame(dst=node.address, payload=b"NORMAL_APP_DATA")
                print(",", end="", flush=True)
        statistics = [
            (f"{node.address:016X}", node.stats.received_rssi_dbm()) for node in mari_edge.nodes
        ]
        print(f"Stats: {statistics}")
        time.sleep(3)


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod

from marilib.marilib import MarilibBase


class MarilibTUI(ABC):
    @abstractmethod
    def render(self, mari: MarilibBase):
        pass

    @abstractmethod
    def close(self):
        pass



# SPDX-FileCopyrightText: 2022-present Inria
# SPDX-FileCopyrightText: 2022-present Alexandre Abadie <alexandre.abadie@inria.fr>
# SPDX-FileCopyrightText: 2025-present Geovane Fedrecheski <geovane.fedrecheski@inria.fr>
#
# SPDX-License-Identifier: BSD-3-Clause

"""Serial interface."""

import logging
import sys
import threading
import time
from typing import Callable

import serial
from serial.tools import list_ports

SERIAL_PAYLOAD_CHUNK_SIZE = 64
SERIAL_PAYLOAD_CHUNK_SIZE_WITH_TRIGGER_BYTE = 63
SERIAL_PAYLOAD_CHUNK_DELAY = 0.003  # 2 ms
SERIAL_DEFAULT_PORT = "/dev/ttyACM0"
SERIAL_DEFAULT_BAUDRATE = 1_000_000
# SERIAL_DEFAULT_BAUDRATE = 460_800


def get_default_port():
    """Return default serial port."""
    ports = [port for port in list_ports.comports()]
    if sys.platform != "win32":
        ports = sorted([port for port in ports if "J-Link" == port.product])
    if not ports:
        return SERIAL_DEFAULT_PORT
    # return first JLink port available
    return ports[0].device


class SerialInterfaceException(Exception):
    """Exception raised when serial port is disconnected."""


class SerialInterface(threading.Thread):
    """Bidirectional serial interface."""

    def __init__(self, port: str, baudrate: int, callback: Callable):
        self.lock = threading.Lock()
        self.callback = callback
        self.serial = serial.Serial(port, baudrate)
        super().__init__(daemon=True)
        self._logger = logging.getLogger(__name__)
        self.start()
        self._logger.info("Serial port thread started")

    def run(self):
        """Listen continuously at each byte received on serial."""
        self.serial.flush()
        try:
            while 1:
                try:
                    byte = self.serial.read(1)
                except (TypeError, serial.serialutil.SerialException):
                    byte = None
                if byte is None:
                    self._logger.info("Serial port disconnected")
                    break
                self.callback(byte)
        except serial.serialutil.PortNotOpenError as exc:
            self._logger.error(f"{exc}")
            raise SerialInterfaceException(f"{exc}") from exc
        except serial.serialutil.SerialException as exc:
            self._logger.error(f"{exc}")
            raise SerialInterfaceException(f"{exc}") from exc

    def stop(self):
        self.serial.close()
        self.join()

    def write_chunked(self, bytes_):
        """Write bytes on serial using the chunked strategy. (deprecated)"""
        # Send 64 bytes at a time
        pos = 0
        while pos < len(bytes_):
            chunk_end = min(pos + SERIAL_PAYLOAD_CHUNK_SIZE, len(bytes_))
            self.serial.write(bytes_[pos:chunk_end])
            self.serial.flush()
            pos = chunk_end
            if pos < len(bytes_):  # Only sleep if there are more chunks
                time.sleep(SERIAL_PAYLOAD_CHUNK_DELAY)

    def write_trigger_byte(self, bytes_):
        """Write bytes on serial using the trigger byte strategy. (deprecated)"""
        self.serial.write(bytes_[0:1])
        time.sleep(
            0.0001
        )  # 100 us -- this time is important because of the trigger byte timeout on the nRF side
        self.serial.write(bytes_[1:])
        self.serial.flush()
        time.sleep(SERIAL_PAYLOAD_CHUNK_DELAY)

    def write_chunked_with_trigger_byte(self, bytes_):
        """Write bytes on serial using the chunked strategy with trigger byte."""
        # Send 64 bytes at a time
        pos = 0
        while pos < len(bytes_):
            # send trigger byte
            trigger_byte = bytes_[pos : pos + 1]
            self.serial.write(trigger_byte)
            time.sleep(
                0.0001
            )  # 100 us -- this time is important because of the trigger byte timeout on the nRF side
            pos += 1

            # send chunk
            chunk_end = min(pos + SERIAL_PAYLOAD_CHUNK_SIZE_WITH_TRIGGER_BYTE, len(bytes_))
            self.serial.write(bytes_[pos:chunk_end])
            self.serial.flush()
            pos = chunk_end

            # sleep to force a delay between chunks (or between calls to this function)
            time.sleep(SERIAL_PAYLOAD_CHUNK_DELAY)

    def write(self, bytes_):
        """Write bytes on serial."""
        self.write_chunked_with_trigger_byte(bytes_)



import threading
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Callable
from rich import print

from marilib.metrics import MetricsTester
from marilib.mari_protocol import (
    MARI_BROADCAST_ADDRESS,
    Frame,
    Header,
    DefaultPayload,
    DefaultPayloadType,
)
from marilib.model import (
    EdgeEvent,
    GatewayInfo,
    MariGateway,
    MariNode,
    NodeInfoEdge,
    SCHEDULES,
)
from marilib.protocol import ProtocolPayloadParserException
from marilib.communication_adapter import MQTTAdapter, MQTTAdapterDummy, SerialAdapter
from marilib.marilib import MarilibBase
from marilib.tui_edge import MarilibTUIEdge


@dataclass
class MarilibEdge(MarilibBase):
    """
    The MarilibEdge class runs in either a computer or a raspberry pi.
    It is used to communicate with:
    - a Mari radio gateway (nRF5340) via serial
    - a Mari cloud instance via MQTT (optional)
    """

    cb_application: Callable[[EdgeEvent, MariNode | Frame], None]
    serial_interface: SerialAdapter
    mqtt_interface: MQTTAdapter | None = None
    tui: MarilibTUIEdge | None = None

    logger: Any | None = None
    gateway: MariGateway = field(default_factory=MariGateway)
    lock: threading.Lock = field(default_factory=threading.Lock, repr=False)
    metrics_tester: MetricsTester | None = None
    metrics_probe_period: float | None = None

    started_ts: datetime = field(default_factory=datetime.now)
    last_received_serial_data_ts: datetime = field(default_factory=datetime.now)
    last_received_mqtt_data_ts: datetime = field(default_factory=datetime.now)
    main_file: str | None = None

    def __post_init__(self):
        self.setup_params = {
            "main_file": self.main_file or "unknown",
            "serial_port": self.serial_interface.port,
        }
        if self.mqtt_interface is None:
            self.mqtt_interface = MQTTAdapterDummy()
        self.serial_interface.init(self.on_serial_data_received)
        if self.logger:
            self.logger.log_setup_parameters(self.setup_params)
        self.metrics_tester = MetricsTester(self, self.metrics_probe_period)
        self.metrics_tester.start()

    # ============================ MarilibBase methods =========================

    def update(self):
        with self.lock:
            self.gateway.update()
            if self.logger and self.logger.active:
                self.logger.log_periodic_metrics(self.gateway, self.gateway.nodes)

    @property
    def nodes(self) -> list[MariNode]:
        return self.gateway.nodes

    def add_node(self, address: int, gateway_address: int = None) -> MariNode | None:
        with self.lock:
            return self.gateway.add_node(address)

    def remove_node(self, address: int) -> MariNode | None:
        with self.lock:
            return self.gateway.remove_node(address)

    def send_frame(self, dst: int, payload: bytes):
        """Sends a frame to the gateway via serial."""
        assert self.serial_interface is not None

        mari_frame = Frame(Header(destination=dst), payload=payload)

        with self.lock:
            self.gateway.register_sent_frame(mari_frame)
            if dst == MARI_BROADCAST_ADDRESS:
                for n in self.gateway.nodes:
                    n.register_sent_frame(mari_frame)
            elif n := self.gateway.get_node(dst):
                n.register_sent_frame(mari_frame)

        self.serial_interface.send_data(
            EdgeEvent.to_bytes(EdgeEvent.NODE_DATA) + mari_frame.to_bytes()
        )

    def render_tui(self):
        if self.tui:
            self.tui.render(self)

    def close_tui(self):
        if self.tui:
            self.tui.close()

    # ============================ MarilibEdge methods =========================

    @property
    def uses_mqtt(self) -> bool:
        return not isinstance(self.mqtt_interface, MQTTAdapterDummy)

    @property
    def mqtt_connected(self) -> bool:
        return self.uses_mqtt and self.mqtt_interface.is_ready()

    @property
    def serial_connected(self) -> bool:
        return self.serial_interface is not None

    def get_max_downlink_rate(self) -> float:
        """Calculate the max downlink packets/sec for a given schedule_id."""
        schedule_params = SCHEDULES.get(self.gateway.info.schedule_id)
        if not schedule_params:
            return 0.0
        d_down = schedule_params["d_down"]
        sf_duration_ms = schedule_params["sf_duration"]
        if sf_duration_ms == 0:
            return 0.0
        return d_down / (sf_duration_ms / 1000.0)

    # ============================ Callbacks ===================================

    def on_mqtt_data_received(self, data: bytes):
        """Just forwards the data to the serial interface."""
        if len(data) < 1:
            return

        self.last_received_mqtt_data_ts = datetime.now()

        try:
            event_type = EdgeEvent(data[0])
            frame = Frame().from_bytes(data[1:])
        except (ValueError, ProtocolPayloadParserException) as exc:
            print(f"[red]Error parsing frame: {exc}[/]")
            return
        if event_type != EdgeEvent.NODE_DATA:
            return
        if frame.header.destination != MARI_BROADCAST_ADDRESS and not self.gateway.get_node(
            frame.header.destination
        ):
            return
        self.send_frame(frame.header.destination, frame.payload)

    def handle_serial_data(self, data: bytes) -> tuple[bool, EdgeEvent, Any]:
        """
        Handles the serial data received from the radio gateway.
        """
        if len(data) < 1:
            return False, EdgeEvent.UNKNOWN, None

        self.last_received_serial_data_ts = datetime.now()

        try:
            event_type = EdgeEvent(data[0])
        except ValueError:
            return False, EdgeEvent.UNKNOWN, None

        if event_type == EdgeEvent.NODE_JOINED:
            node_info = NodeInfoEdge().from_bytes(data[1:])
            self.add_node(node_info.address)
            return True, event_type, node_info

        elif event_type == EdgeEvent.NODE_LEFT:
            node_info = NodeInfoEdge().from_bytes(data[1:])
            if self.remove_node(node_info.address):
                return True, event_type, node_info
            else:
                return False, event_type, node_info

        elif event_type == EdgeEvent.NODE_KEEP_ALIVE:
            node_info = NodeInfoEdge().from_bytes(data[1:])
            with self.lock:
                self.gateway.update_node_liveness(node_info.address)
            return True, event_type, node_info

        elif event_type == EdgeEvent.GATEWAY_INFO:
            try:
                with self.lock:
                    self.gateway.set_info(GatewayInfo().from_bytes(data[1:]))
                return True, event_type, self.gateway.info
            except (ValueError, ProtocolPayloadParserException):
                return False, EdgeEvent.UNKNOWN, None

        elif event_type == EdgeEvent.NODE_DATA:
            try:
                frame = Frame().from_bytes(data[1:])
                with self.lock:
                    self.gateway.update_node_liveness(frame.header.source)
                    self.gateway.register_received_frame(frame)

                    # handle metrics probe packets
                    if frame.is_test_packet:
                        payload = self.metrics_tester.handle_response_edge(frame)
                        if payload:
                            frame.payload = payload.to_bytes()

                return True, event_type, frame
            except (ValueError, ProtocolPayloadParserException):
                return False, EdgeEvent.UNKNOWN, None

        return False, event_type, None

    def on_serial_data_received(self, data: bytes):
        res, event_type, event_data = self.handle_serial_data(data)
        if not res:
            return

        if self.logger and event_type in [EdgeEvent.NODE_JOINED, EdgeEvent.NODE_LEFT]:
            self.logger.log_event(self.gateway.info.address, event_data.address, event_type.name)
        if event_type == EdgeEvent.GATEWAY_INFO:
            self.mqtt_interface.update(event_data.network_id_str, self.on_mqtt_data_received)
            if self.logger:
                self.setup_params["schedule_name"] = self.gateway.info.schedule_name
                self.logger.log_setup_parameters(self.setup_params)

        if event_type == EdgeEvent.NODE_DATA and not event_data.is_test_packet:
            # only notify the application if it's not a test packet
            self.cb_application(event_type, event_data)

        self.send_data_to_cloud(event_type, event_data)

    def send_data_to_cloud(
        self, event_type: EdgeEvent, event_data: NodeInfoEdge | GatewayInfo | Frame
    ):
        if event_type in [EdgeEvent.NODE_JOINED, EdgeEvent.NODE_LEFT, EdgeEvent.NODE_KEEP_ALIVE]:
            event_data = event_data.to_cloud(self.gateway.info.address)
        data = EdgeEvent.to_bytes(event_type) + event_data.to_bytes()
        self.mqtt_interface.send_data_to_cloud(data)

    # ============================ Utility methods =============================

    # def metrics_test_enable(self):
    #     self.metrics_tester.start()

    def metrics_test_disable(self):
        self.metrics_tester.stop()

    # ============================ Private methods =============================

    def _is_test_packet(self, payload: bytes) -> bool:
        """Determines if a packet sent FROM the edge is for testing purposes."""
        payload = DefaultPayload().from_bytes(payload)
        return payload.type_ in [DefaultPayloadType.LOAD_TEST, DefaultPayloadType.LATENCY_TEST]



import time

import click
from marilib.logger import MetricsLogger
from marilib.mari_protocol import Frame, MARI_BROADCAST_ADDRESS, DefaultPayload
from marilib.model import EdgeEvent, MariNode
from marilib.communication_adapter import SerialAdapter, MQTTAdapter
from marilib.serial_uart import get_default_port
from marilib.tui_edge import MarilibTUIEdge
from marilib.marilib_edge import MarilibEdge


def on_event(event: EdgeEvent, event_data: MariNode | Frame):
    """An event handler for the application."""
    pass


@click.command()
@click.option(
    "--port",
    "-p",
    type=str,
    default=get_default_port(),
    show_default=True,
    help="Serial port to use (e.g., /dev/ttyACM0)",
)
@click.option(
    "--mqtt-url",
    "-m",
    type=str,
    default=None,
    help="MQTT broker to use (default: None, no cloud)",
)
@click.option(
    "--metrics-probe-interval",
    "-i",
    type=float,
    default=0,
    help="How often to send a metrics probe in seconds (default: 0, no metrics)",
)
@click.option(
    "--log-dir",
    default="logs",
    show_default=True,
    help="Directory to save metric log files.",
    type=click.Path(),
)
def main(port: str | None, mqtt_url: str, metrics_probe_interval: float, log_dir: str):
    """A basic example of using the MarilibEdge library."""

    mari = MarilibEdge(
        on_event,
        serial_interface=SerialAdapter(port),
        mqtt_interface=MQTTAdapter.from_url(mqtt_url, is_edge=True) if mqtt_url else None,
        logger=MetricsLogger(
            log_dir_base=log_dir, rotation_interval_minutes=1440, log_interval_seconds=1.0
        ),
        tui=MarilibTUIEdge(),
        main_file=__file__,
        metrics_probe_period=metrics_probe_interval,  # use a less frequent probe to interfere less with the main traffic
    )

    try:
        while True:
            mari.update()
            if not mari.uses_mqtt and mari.nodes:
                mari.send_frame(MARI_BROADCAST_ADDRESS, DefaultPayload().to_bytes())
            mari.render_tui()
            time.sleep(0.5)
    except KeyboardInterrupt:
        pass
    finally:
        mari.close_tui()
        mari.logger.close()


if __name__ == "__main__":
    main()



import threading
from typing import TYPE_CHECKING
from rich import print

from marilib.model import NodeStatsReply
from marilib.mari_protocol import Frame
from marilib.protocol import ProtocolPayloadParserException


if TYPE_CHECKING:
    from marilib.marilib_edge import MarilibEdge


PDR_STATS_REQUEST_PAYLOAD = b"S"


class PDRTester:
    """A thread-based class to periodically test PDR to all nodes."""

    def __init__(self, marilib: "MarilibEdge", interval: float = 15.0):
        self.marilib = marilib
        self.interval = interval
        self._stop_event = threading.Event()
        self._thread = threading.Thread(target=self._run, daemon=True)

    def start(self):
        """Starts the PDR testing thread."""
        print("[yellow]PDR tester started.[/]")
        self._thread.start()

    def stop(self):
        """Stops the PDR testing thread."""
        self._stop_event.set()
        if self._thread.is_alive():
            self._thread.join()
        print("[yellow]PDR tester stopped.[/]")

    def _run(self):
        """The main loop for the testing thread."""
        self._stop_event.wait(self.interval)

        while not self._stop_event.is_set():
            nodes = list(self.marilib.nodes)
            if not nodes:
                self._stop_event.wait(self.interval)
                continue

            for node in nodes:
                if self._stop_event.is_set():
                    break
                self.send_pdr_request(node.address)

                # Spread requests evenly over the interval
                sleep_duration = self.interval / len(nodes)
                self._stop_event.wait(sleep_duration)

    def send_pdr_request(self, address: int):
        """Sends a PDR stats request to a specific address."""
        self.marilib.send_frame(address, PDR_STATS_REQUEST_PAYLOAD)

    def handle_response(self, frame: Frame) -> bool:
        """
        Handles a PDR stats response frame and calculates PDR values.
        Returns True if the frame was a valid stats reply, False otherwise.
        """
        if len(frame.payload) != 8:
            return False

        try:
            stats_reply = NodeStatsReply().from_bytes(frame.payload)
            node = self.marilib.gateway.get_node(frame.header.source)

            if node:
                # Update with the latest stats reported by the node
                node.last_reported_rx_count = stats_reply.rx_app_packets
                node.last_reported_tx_count = stats_reply.tx_app_packets

                # Calculate Downlink PDR
                sent_count = node.stats.sent_count(include_test_packets=False)
                if sent_count > 0:
                    pdr = node.last_reported_rx_count / sent_count
                    node.pdr_downlink = min(pdr, 1.0)
                else:
                    node.pdr_downlink = 0.0

                # Calculate Uplink PDR
                received_count = node.stats.received_count(include_test_packets=False)
                if node.last_reported_tx_count > 0:
                    pdr = received_count / node.last_reported_tx_count
                    node.pdr_uplink = min(pdr, 1.0)
                else:
                    node.pdr_uplink = 0.0

                return True

        except (ValueError, ProtocolPayloadParserException):
            return False

        return False



import base64
from urllib.parse import urlparse
import paho.mqtt.client as mqtt

from abc import ABC, abstractmethod
from rich import print

from marilib.serial_hdlc import (
    HDLCDecodeException,
    HDLCHandler,
    HDLCState,
    hdlc_encode,
)
from marilib.serial_uart import SerialInterface, SERIAL_DEFAULT_BAUDRATE


class CommunicationAdapterBase(ABC):
    """Base class for interface adapters."""

    @abstractmethod
    def init(self, on_data_received: callable):
        """Initialize the interface."""

    @abstractmethod
    def close(self):
        """Close the interface."""


class SerialAdapter(CommunicationAdapterBase):
    """Class used to interface with the serial port."""

    def __init__(self, port, baudrate=SERIAL_DEFAULT_BAUDRATE):
        self.port = port
        self.baudrate = baudrate
        self.hdlc_handler = HDLCHandler()

    def on_byte_received(self, byte):
        self.hdlc_handler.handle_byte(byte)
        if self.hdlc_handler.state == HDLCState.READY:
            try:
                payload = self.hdlc_handler.payload
                # print(f"Received payload: {payload.hex()}")
                self.on_data_received(payload)
            except HDLCDecodeException as e:
                print(f"Error decoding payload: {e}")

    def init(self, on_data_received: callable):
        self.on_data_received = on_data_received
        self.serial = SerialInterface(self.port, self.baudrate, self.on_byte_received)
        print(f"[yellow]Connected to serial port {self.port} at {self.baudrate} baud[/]")

    def close(self):
        print("[yellow]Disconnect from gateway...[/]")

    def send_data(self, data):
        with self.serial.lock:  # Use the existing lock for thread safety
            self.serial.serial.flush()
            encoded = hdlc_encode(data)
            self.serial.write(encoded)


class MQTTAdapter(CommunicationAdapterBase):
    """Class used to interface with MQTT."""

    def __init__(self, host, port, is_edge: bool, use_tls: bool = False):
        self.host = host
        self.port = port
        self.is_edge = is_edge
        self.network_id = None
        self.client = None
        self.on_data_received = None
        self.use_tls = use_tls
        # optimize qos for throughput
        # 0 = no delivery guarantee, 1 = at least once, 2 = exactly once
        self.qos = 0

    @classmethod
    def from_url(cls, url: str, is_edge: bool):
        url = urlparse(url)
        host, port = url.netloc.split(":")
        if url.scheme == "mqtt":
            return cls(host, int(port), is_edge, use_tls=False)
        elif url.scheme == "mqtts":
            return cls(host, int(port), is_edge, use_tls=True)
        else:
            raise ValueError(f"Invalid MQTT URL: {url} (must start with mqtt:// or mqtts://)")

    # ==== public methods ====

    def is_ready(self) -> bool:
        return self.client is not None and self.client.is_connected()

    def set_network_id(self, network_id: str):
        self.network_id = network_id

    def set_on_data_received(self, on_data_received: callable):
        self.on_data_received = on_data_received

    def update(self, network_id: str, on_data_received: callable):
        if self.network_id is None:
            # might have been set by set_network_id
            self.network_id = network_id
        else:
            # TODO: handle the case when the network_id changes
            pass
        if self.network_id is None:
            # wait a bit, network_id not set yet
            return
        if self.on_data_received is None:
            self.set_on_data_received(on_data_received)
        if not self.is_ready():
            self.init()

    def init(self):
        if self.client:
            # already initialized, do nothing
            return
        if self.network_id is None:
            # network_id not set yet
            return

        self.client = mqtt.Client(
            mqtt.CallbackAPIVersion.VERSION2,
            protocol=mqtt.MQTTProtocolVersion.MQTTv5,
        )
        if self.use_tls:
            self.client.tls_set_context(context=None)
        self.client.on_log = self._on_log
        self.client.on_connect = self._on_connect_edge if self.is_edge else self._on_connect_cloud
        self.client.on_message = self._on_message_edge if self.is_edge else self._on_message_cloud
        self.client.connect(self.host, self.port, 60)
        print(f"[yellow]Connected to MQTT broker on {self.host}:{self.port}[/]")
        self.client.loop_start()

    def close(self):
        self.client.disconnect()
        self.client.loop_stop()

    def send_data_to_edge(self, data):
        if not self.is_ready():
            return
        self.client.publish(
            f"/mari/{self.network_id}/to_edge",
            base64.b64encode(data).decode(),
            qos=self.qos,
        )

    def send_data_to_cloud(self, data):
        if not self.is_ready():
            return
        self.client.publish(
            f"/mari/{self.network_id}/to_cloud",
            base64.b64encode(data).decode(),
            qos=self.qos,
        )

    # ==== private methods ====

    # TODO: de-duplicate the _on_message_* functions? decide as the integration evolves
    def _on_message_edge(self, client, userdata, message):
        try:
            data = base64.b64decode(message.payload)
        except Exception as e:
            # print the error and a stacktrace
            print(f"[red]Error decoding MQTT message: {e}[/]")
            print(f"[red]Message: {message.payload}[/]")
            return
        self.on_data_received(data)

    def _on_message_cloud(self, client, userdata, message):
        try:
            data = base64.b64decode(message.payload)
        except Exception as e:
            # print the error and a stacktrace
            print(f"[red]Error decoding MQTT message: {e}[/]")
            print(f"[red]Message: {message.payload}[/]")
            return
        self.on_data_received(data)

    def _on_log(self, client, userdata, paho_log_level, messages):
        # print(messages)
        pass

    def _on_connect_edge(self, client, userdata, flags, reason_code, properties):
        self.client.subscribe(f"/mari/{self.network_id}/to_edge", qos=self.qos)
        print(f"[yellow]Subscribed to /mari/{self.network_id}/to_edge[/]")

    def _on_connect_cloud(self, client, userdata, flags, reason_code, properties):
        self.client.subscribe(f"/mari/{self.network_id}/to_cloud", qos=self.qos)
        print(f"[yellow]Subscribed to /mari/{self.network_id}/to_cloud[/]")


class MQTTAdapterDummy(MQTTAdapter):
    """Dummy MQTT adapter, does nothing, for when edge runs only locally, without a cloud."""

    def __init__(self, host="", port=0, is_edge=True):
        super().__init__(host, port, is_edge)

    def is_ready(self) -> bool:
        """Dummy adapter is never ready."""
        return False

    def init(self):
        pass

    def close(self):
        pass

    def send_data_to_edge(self, data):
        pass

    def send_data_to_cloud(self, data):
        pass



import dataclasses
from dataclasses import dataclass
from enum import IntEnum

from marilib.protocol import Packet, PacketFieldMetadata, PacketType

MARI_PROTOCOL_VERSION = 2
MARI_BROADCAST_ADDRESS = 0xFFFFFFFFFFFFFFFF
MARI_NET_ID_DEFAULT = 0x0001


class DefaultPayloadType(IntEnum):
    APPLICATION_DATA = 1
    METRICS_REQUEST = 128
    METRICS_RESPONSE = 129
    METRICS_LOAD = 130
    METRICS_PROBE = 140

    def as_bytes(self) -> bytes:
        return bytes([self.value])


@dataclass
class DefaultPayload(Packet):
    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="type", length=1),
        ]
    )
    type_: DefaultPayloadType = DefaultPayloadType.APPLICATION_DATA

    def with_filler_bytes(self, length: int) -> bytes:
        return self.to_bytes() + bytes([0xF1] * length)


@dataclass
class MetricsProbePayload(Packet):
    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="type", length=1),
            PacketFieldMetadata(name="cloud_tx_ts_us", length=8),
            PacketFieldMetadata(name="cloud_rx_ts_us", length=8),
            PacketFieldMetadata(name="cloud_tx_count", length=4),
            PacketFieldMetadata(name="cloud_rx_count", length=4),
            PacketFieldMetadata(name="edge_tx_ts_us", length=8),
            PacketFieldMetadata(name="edge_rx_ts_us", length=8),
            PacketFieldMetadata(name="edge_tx_count", length=4),
            PacketFieldMetadata(name="edge_rx_count", length=4),
            PacketFieldMetadata(name="gw_tx_count", length=4),
            PacketFieldMetadata(name="gw_rx_count", length=4),
            PacketFieldMetadata(name="gw_rx_asn", length=8),
            PacketFieldMetadata(name="gw_tx_enqueued_asn", length=8),
            PacketFieldMetadata(name="gw_tx_dequeued_asn", length=8),
            PacketFieldMetadata(name="node_rx_count", length=4),
            PacketFieldMetadata(name="node_tx_count", length=4),
            PacketFieldMetadata(name="node_rx_asn", length=8),
            PacketFieldMetadata(name="node_tx_enqueued_asn", length=8),
            PacketFieldMetadata(name="node_tx_dequeued_asn", length=8),
            PacketFieldMetadata(name="rssi_at_node", length=1),
            PacketFieldMetadata(name="rssi_at_gw", length=1),
        ]
    )
    type_: DefaultPayloadType = DefaultPayloadType.METRICS_PROBE
    cloud_tx_ts_us: int = 0
    cloud_rx_ts_us: int = 0
    cloud_tx_count: int = 0
    cloud_rx_count: int = 0
    edge_tx_ts_us: int = 0
    edge_rx_ts_us: int = 0
    edge_tx_count: int = 0
    edge_rx_count: int = 0
    gw_tx_count: int = 0
    gw_rx_count: int = 0
    gw_rx_asn: int = 0
    gw_tx_enqueued_asn: int = 0
    gw_tx_dequeued_asn: int = 0
    node_rx_count: int = 0
    node_tx_count: int = 0
    node_rx_asn: int = 0
    node_tx_enqueued_asn: int = 0
    node_tx_dequeued_asn: int = 0
    rssi_at_node: int = 0
    rssi_at_gw: int = 0

    @property
    def packet_length(self) -> int:
        return sum(field.length for field in self.metadata)

    @property
    def asn(self) -> int:
        """ASN at reception back from the network"""
        return self.gw_rx_asn

    def latency_roundtrip_node_edge_ms(self) -> float:
        return (self.edge_rx_ts_us - self.edge_tx_ts_us) / 1000.0

    def latency_roundtrip_node_cloud_ms(self) -> float:
        return (self.cloud_rx_ts_us - self.cloud_tx_ts_us) / 1000.0

    def pdr_saturated(self, count_a: int, count_b: int) -> float:
        if count_b == 0:
            return 0.0
        pdr = count_a / count_b
        if pdr > 1.0:
            return 0.0
        return pdr

    def pdr_uplink_radio(self, probe_stats_start_epoch=None) -> float:
        if probe_stats_start_epoch is None:
            # if no epoch is provided, use the current values
            gw_rx_count = self.gw_rx_count
            node_tx_count = self.node_tx_count
        else:
            # if epoch is provided, subtract the epoch values from the current values
            if probe_stats_start_epoch.asn == 0:
                return 0
            gw_rx_count = self.gw_rx_count - probe_stats_start_epoch.gw_rx_count
            node_tx_count = self.node_tx_count - probe_stats_start_epoch.node_tx_count
        if node_tx_count <= 0:
            return 0
        return self.pdr_saturated(gw_rx_count, node_tx_count)

    def pdr_downlink_radio(self, probe_stats_start_epoch=None) -> float:
        if probe_stats_start_epoch is None:
            # if no epoch is provided, use the current values
            gw_tx_count = self.gw_tx_count
            node_rx_count = self.node_rx_count
        else:
            # if epoch is provided, subtract the epoch values from the current values
            if probe_stats_start_epoch.asn == 0:
                return 0
            gw_tx_count = self.gw_tx_count - probe_stats_start_epoch.gw_tx_count
            node_rx_count = self.node_rx_count - probe_stats_start_epoch.node_rx_count
        return self.pdr_saturated(node_rx_count, gw_tx_count)

    def pdr_uplink_uart(self, probe_stats_start_epoch=None) -> float:
        if probe_stats_start_epoch is None:
            # if no epoch is provided, just wait a bit
            return -1
        else:
            # if epoch is provided, subtract the epoch values from the current values
            if probe_stats_start_epoch.asn == 0:
                return 0
            # if a packet was received at the gatweway, it should also be received at the edge (otherwise, it's a loss)
            gw_rx_count = self.gw_rx_count - probe_stats_start_epoch.gw_rx_count
            edge_rx_count = self.edge_rx_count - probe_stats_start_epoch.edge_rx_count
        return self.pdr_saturated(edge_rx_count, gw_rx_count)

    def pdr_downlink_uart(self, probe_stats_start_epoch=None) -> float:
        if probe_stats_start_epoch is None:
            # if no epoch is provided, just wait a bit
            return -1
        else:
            # if epoch is provided, subtract the epoch values from the current values
            if probe_stats_start_epoch.asn == 0:
                return 0
            # if a packet was sent at the edge, it should also be sent at the gateway (otherwise, it's a loss)
            gw_tx_count = self.gw_tx_count - probe_stats_start_epoch.gw_tx_count
            edge_tx_count = self.edge_tx_count - probe_stats_start_epoch.edge_tx_count
        return self.pdr_saturated(gw_tx_count, edge_tx_count)

    def rssi_at_node_dbm(self) -> int:
        if self.rssi_at_node > 127:
            return self.rssi_at_node - 255
        return self.rssi_at_node

    def rssi_at_gw_dbm(self) -> int:
        if self.rssi_at_gw > 127:
            return self.rssi_at_gw - 255
        return self.rssi_at_gw

    def __repr__(self):
        rep = dataclasses.asdict(self)
        rep.pop("metadata", None)
        return f"{rep}"


@dataclass
class MetricsRequestPayload(Packet):
    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="type", length=1),
            PacketFieldMetadata(name="timestamp_us", length=8),
        ]
    )
    type_: DefaultPayloadType = DefaultPayloadType.METRICS_REQUEST
    timestamp_us: int = 0


@dataclass
class MetricsResponsePayload(Packet):
    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="type", length=1),
            PacketFieldMetadata(name="timestamp_us", length=8),
            PacketFieldMetadata(name="rx_count", length=4),
            PacketFieldMetadata(name="tx_count", length=4),
        ]
    )
    type_: DefaultPayloadType = DefaultPayloadType.METRICS_RESPONSE
    timestamp_us: int = 0
    rx_count: int = 0
    tx_count: int = 0


@dataclass
class HeaderStats(Packet):
    """Dataclass that holds MAC header stats."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="rssi", disp="rssi", length=1),
        ]
    )
    rssi: int = 0

    @property
    def rssi_dbm(self) -> int:
        if self.rssi > 127:
            return self.rssi - 255
        return self.rssi


@dataclass
class Header(Packet):
    """Dataclass that holds MAC header fields."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="version", disp="ver.", length=1),
            PacketFieldMetadata(name="type_", disp="type", length=1),
            PacketFieldMetadata(name="network_id", disp="net", length=2),
            PacketFieldMetadata(name="destination", disp="dst", length=8),
            PacketFieldMetadata(name="source", disp="src", length=8),
        ]
    )
    version: int = MARI_PROTOCOL_VERSION
    type_: int = PacketType.DATA
    network_id: int = MARI_NET_ID_DEFAULT
    destination: int = MARI_BROADCAST_ADDRESS
    source: int = 0x0000000000000000

    def __repr__(self):
        type_ = PacketType(self.type_).name
        return f"Header(version={self.version}, type_={type_}, network_id=0x{self.network_id:04x}, destination=0x{self.destination:016x}, source=0x{self.source:016x})"


@dataclass
class Frame:
    """Data class that holds a payload packet."""

    header: Header = None
    stats: HeaderStats = dataclasses.field(default_factory=HeaderStats)
    payload: bytes = b""

    def from_bytes(self, bytes_):
        self.header = Header().from_bytes(bytes_[0:20])
        if len(bytes_) > 20:
            self.stats = HeaderStats().from_bytes(bytes_[20:21])
            if len(bytes_) > 21:
                self.payload = bytes_[21:]
        return self

    def to_bytes(self, byteorder="little") -> bytes:
        header_bytes = self.header.to_bytes(byteorder)
        stats_bytes = self.stats.to_bytes(byteorder)
        return header_bytes + stats_bytes + self.payload

    @property
    def is_test_packet(self) -> bool:
        """Returns True if either the payload is a metrics response, request, or load test packet."""
        return (
            self.payload.startswith(DefaultPayloadType.METRICS_RESPONSE.as_bytes())
            or self.payload.startswith(DefaultPayloadType.METRICS_REQUEST.as_bytes())
            or self.payload.startswith(DefaultPayloadType.METRICS_LOAD.as_bytes())
            or self.payload.startswith(DefaultPayloadType.METRICS_PROBE.as_bytes())
        )

    @property
    def is_load_test_packet(self) -> bool:
        return self.payload.startswith(DefaultPayloadType.METRICS_LOAD.as_bytes())

    def __repr__(self):
        header_no_metadata = dataclasses.replace(self.header, metadata=[])
        return f"Frame(header={header_no_metadata}, payload={self.payload})"



import time

from marilib.serial_hdlc import (
    HDLCDecodeException,
    HDLCHandler,
    HDLCState,
    hdlc_encode,
)
from marilib.serial_uart import SerialInterface

BAUDRATE = 1000000

hdlc_handler = HDLCHandler()


def on_byte_received(byte):
    hdlc_handler.handle_byte(byte)
    if hdlc_handler.state == HDLCState.READY:
        try:
            print(".", end="", flush=True)
        except HDLCDecodeException as e:
            print(f"Error decoding payload: {e}")


serial_interface = SerialInterface("/dev/ttyACM0", BAUDRATE, on_byte_received)


while True:

    def send_payload(payload):
        print(f"\nSending {len(payload)} bytes: {payload.hex(' ')}")
        encoded = hdlc_encode(payload)
        print(f"Sending encoded {len(encoded)} bytes: {encoded.hex(' ')}")
        serial_interface.write_chunked_with_trigger_byte(encoded)

    sleep_time = 0.01  # 10 ms

    for i in range(4):
        # send_payload(b"ABCD" * 8) # 32 bytes
        send_payload(b"A" * 80)
        time.sleep(sleep_time)

    print("Sleeping for 100 ms\n\n")
    time.sleep(0.1)



import csv
import os
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import IO, List, Dict

from marilib.model import MariGateway, MariNode


@dataclass
class MetricsLogger:
    """
    A metrics logger that saves statistics to CSV files with log rotation.
    """

    log_dir_base: str = "logs"
    rotation_interval_minutes: int = 1440  # 1 day
    already_logged_setup_parameters: bool = False
    log_interval_seconds: float = 1.0
    last_log_time: Dict[int, datetime] = field(default_factory=dict)

    def __post_init__(self):
        """
        Initializes the logger with rotation and setup logging capabilities.
        """
        try:
            self.rotation_interval = timedelta(minutes=self.rotation_interval_minutes)

            self.start_time = datetime.now()
            self.run_timestamp = self.start_time.strftime("%Y%m%d_%H%M%S")
            self.log_dir = os.path.join(self.log_dir_base, f"run_{self.run_timestamp}")
            os.makedirs(self.log_dir, exist_ok=True)

            self._gateway_file: IO[str] | None = None
            self._nodes_file: IO[str] | None = None
            self._events_file: IO[str] | None = None
            self._gateway_writer = None
            self._nodes_writer = None
            self._events_writer = None
            self.segment_start_time: datetime | None = None

            # Open events log file
            events_path = os.path.join(self.log_dir, "log_events.csv")
            self._events_file = open(events_path, "w", newline="", encoding="utf-8")
            self._events_writer = csv.writer(self._events_file)
            self._events_writer.writerow(
                ["timestamp", "gateway_address", "node_address", "event_name", "event_tag"]
            )

            self._open_new_segment()
            self.active = True

        except (IOError, OSError) as e:
            print(f"Error: Failed to initialize logger: {e}")
            self.active = False

    def log_setup_parameters(self, params: Dict[str, any] | None):
        """Creates and writes test setup parameters to metrics_setup.csv."""
        if not params or self.already_logged_setup_parameters:
            return
        # only log setup parameters once
        self.already_logged_setup_parameters = True

        setup_path = os.path.join(self.log_dir, "metrics_setup.csv")
        with open(setup_path, "w", newline="", encoding="utf-8") as f:
            writer = csv.writer(f)
            writer.writerow(["param", "value"])
            writer.writerow(["start_time", self.start_time.isoformat()])
            for key, value in params.items():
                writer.writerow([key, value])

    def _open_new_segment(self):
        self._close_segment_files()

        self.segment_start_time = datetime.now()
        segment_ts = self.segment_start_time.strftime("%H%M%S")

        gateway_path = os.path.join(self.log_dir, f"gateway_metrics_{segment_ts}.csv")
        nodes_path = os.path.join(self.log_dir, f"node_metrics_{segment_ts}.csv")

        self._gateway_file = open(gateway_path, "w", newline="", encoding="utf-8")
        self._gateway_writer = csv.writer(self._gateway_file)
        gateway_header = [
            "timestamp",
            "gateway_address",
            "schedule_id",
            "connected_nodes",
            # "tx_total",
            # "rx_total",
            # "tx_rate_1s",
            # "rx_rate_1s",
            "avg_latency_ms",
            "avg_pdr_downlink_radio",
            "avg_pdr_uplink_radio",
            "latest_node_tx_count",
            "latest_node_rx_count",
            "latest_gw_tx_count",
            "latest_gw_rx_count",
        ]
        self._gateway_writer.writerow(gateway_header)

        self._nodes_file = open(nodes_path, "w", newline="", encoding="utf-8")
        self._nodes_writer = csv.writer(self._nodes_file)
        nodes_header = [
            "timestamp",
            "gateway_address",
            "node_address",
            "is_alive",
            # "tx_total",
            # "rx_total",
            # "tx_rate_1s",
            # "rx_rate_1s",
            "success_rate_30s",
            "success_rate_total",
            "pdr_downlink",
            "pdr_uplink",
            "rssi_node_dbm",
            "rssi_gw_dbm",
            "avg_latency_edge_ms",
            "avg_latency_cloud_ms",
            "last_latency_edge_ms",
            "last_latency_cloud_ms",
        ]
        self._nodes_writer.writerow(nodes_header)

    def _check_for_rotation(self):
        if datetime.now() - self.segment_start_time >= self.rotation_interval:
            self._open_new_segment()

    def _log_common(self):
        if not self.active:
            return False
        self._check_for_rotation()
        return True

    def log_periodic_metrics(self, gateway: MariGateway, nodes: List[MariNode]):
        last_log_time = self.last_log_time.get(gateway.info.address, self.segment_start_time)
        if datetime.now() - last_log_time >= timedelta(seconds=self.log_interval_seconds):
            self.log_gateway_metrics(gateway)
            self.log_all_nodes_metrics(nodes)
            self.last_log_time[gateway.info.address] = datetime.now()

    def log_gateway_metrics(self, gateway: MariGateway):
        if not self._log_common() or self._gateway_writer is None:
            return

        timestamp = datetime.now().isoformat()
        row = [
            timestamp,
            f"0x{gateway.info.address:016X}",
            gateway.info.schedule_id,
            len(gateway.nodes),
            # gateway.stats.sent_count(include_test_packets=False),
            # gateway.stats.received_count(include_test_packets=False),
            # gateway.stats.sent_count(1, include_test_packets=False),
            # gateway.stats.received_count(1, include_test_packets=False),
            f"{gateway.stats_avg_latency_roundtrip_node_edge_ms():.2f}",
            f"{gateway.stats_avg_pdr_downlink_radio():.2f}",
            f"{gateway.stats_avg_pdr_uplink_radio():.2f}",
            gateway.stats_latest_node_tx_count(),
            gateway.stats_latest_node_rx_count(),
            gateway.stats_latest_gw_tx_count(),
            gateway.stats_latest_gw_rx_count(),
        ]
        self._gateway_writer.writerow(row)

    def log_all_nodes_metrics(self, nodes: List[MariNode]):
        """Writes metrics for all nodes, handling rotation."""
        if not self._log_common() or self._nodes_writer is None:
            return

        timestamp = datetime.now().isoformat()
        for node in nodes:
            row = [
                timestamp,
                f"0x{node.gateway_address:016X}",
                f"0x{node.address:016X}",
                node.is_alive,
                # node.stats.sent_count(include_test_packets=False),
                # node.stats.received_count(include_test_packets=False),
                # node.stats.sent_count(1, include_test_packets=False),
                # node.stats.received_count(1, include_test_packets=False),
                f"{node.stats.success_rate(30):.2%}",
                f"{node.stats.success_rate():.2%}",
                f"{node.pdr_downlink:.2%}",
                f"{node.pdr_uplink:.2%}",
                node.stats_rssi_node_dbm(),
                node.stats_rssi_gw_dbm(),
                f"{node.stats_avg_latency_roundtrip_node_edge_ms():.2f}",
                f"{node.stats_avg_latency_roundtrip_node_edge_ms():.2f}",  # FIXME!: should use cloud option
                f"{node.stats_latest_latency_roundtrip_node_edge_ms():.2f}",
                f"{node.stats_latest_latency_roundtrip_node_edge_ms():.2f}",  # FIXME!: should use cloud option
            ]
            self._nodes_writer.writerow(row)

    def log_event(
        self, gateway_address: int, node_address: int, event_name: str, event_tag: str = ""
    ):
        """Logs an event to the events log file."""
        if not self.active or self._events_writer is None:
            return

        timestamp = datetime.now().isoformat()
        row = [
            timestamp,
            f"0x{gateway_address:016X}",
            f"0x{node_address:016X}",
            event_name,
            event_tag,
        ]
        self._events_writer.writerow(row)
        if self._events_file:
            self._events_file.flush()

    def _close_segment_files(self):
        if self._gateway_file and not self._gateway_file.closed:
            self._gateway_file.close()
        if self._nodes_file and not self._nodes_file.closed:
            self._nodes_file.close()

    def close(self):
        if not self.active:
            return

        self._close_segment_files()
        if self._events_file and not self._events_file.closed:
            self._events_file.close()
        print(f"\nMetrics saved to: {self.log_dir}")
        self.active = False



import threading
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Callable

from marilib.metrics import MetricsTester
from marilib.mari_protocol import Frame, Header
from marilib.model import (
    EdgeEvent,
    GatewayInfo,
    MariGateway,
    MariNode,
    NodeInfoCloud,
)
from marilib.communication_adapter import MQTTAdapter
from marilib.marilib import MarilibBase
from marilib.tui_cloud import MarilibTUICloud

LOAD_PACKET_PAYLOAD = b"L"


@dataclass
class MarilibCloud(MarilibBase):
    """
    The MarilibCloud class runs in a computer.
    It is used to communicate with a Mari radio gateway (nRF5340) via MQTT.
    """

    cb_application: Callable[[EdgeEvent, MariNode | Frame | GatewayInfo], None]
    mqtt_interface: MQTTAdapter
    network_id: int
    tui: MarilibTUICloud | None = None

    logger: Any | None = None
    gateways: dict[int, MariGateway] = field(default_factory=dict)
    lock: threading.Lock = field(default_factory=threading.Lock, repr=False)
    metrics_tester: MetricsTester | None = None

    started_ts: datetime = field(default_factory=datetime.now)
    last_received_mqtt_data_ts: datetime = field(default_factory=datetime.now)
    main_file: str | None = None

    def __post_init__(self):
        self.setup_params = {
            "main_file": self.main_file or "unknown",
            "mqtt_host": self.mqtt_interface.host,
            "mqtt_port": self.mqtt_interface.port,
            "network_id": self.network_id_str,
        }
        self.mqtt_interface.set_network_id(self.network_id_str)
        self.mqtt_interface.set_on_data_received(self.on_mqtt_data_received)
        self.mqtt_interface.init()
        if self.logger:
            self.logger.log_setup_parameters(self.setup_params)
        self.metrics_tester = MetricsTester(
            self
        )  # just instantiate, do not start it at the cloud, for now

    # ============================ MarilibBase methods =========================

    def update(self):
        """Recurrent bookkeeping. Don't forget to call this periodically on your main loop."""
        with self.lock:
            # remove dead gateways
            self.gateways = {
                addr: gateway for addr, gateway in self.gateways.items() if gateway.is_alive
            }
            # update each gateway
            for gateway in self.gateways.values():
                gateway.update()
                if self.logger:
                    self.logger.log_periodic_metrics(gateway, gateway.nodes)

    @property
    def nodes(self) -> list[MariNode]:
        return [node for gateway in self.gateways.values() for node in gateway.nodes]

    def add_node(self, address: int, gateway_address: int = None) -> MariNode | None:
        with self.lock:
            gateway = self.gateways.get(gateway_address)
            if gateway:
                node = gateway.add_node(address)
                return node
        return None

    def remove_node(self, address: int, gateway_address: int = None) -> MariNode | None:
        with self.lock:
            gateway = self.gateways.get(gateway_address)
            if gateway:
                node = gateway.remove_node(address)
                return node
        return None

    def send_frame(self, dst: int, payload: bytes):
        """
        Sends a frame to a gateway via MQTT.
        Consists in publishing a message to the /mari/{network_id}/to_edge topic.
        """
        mari_frame = Frame(Header(destination=dst), payload=payload)

        self.mqtt_interface.send_data_to_edge(
            EdgeEvent.to_bytes(EdgeEvent.NODE_DATA) + mari_frame.to_bytes()
        )

    def render_tui(self):
        if self.tui:
            self.tui.render(self)

    def close_tui(self):
        if self.tui:
            self.tui.close()

    # ============================ MarilibCloud methods =========================

    @property
    def network_id_str(self) -> str:
        return f"{self.network_id:04X}"

    # ============================ Callbacks ===================================

    def handle_mqtt_data(self, data: bytes) -> tuple[bool, EdgeEvent, Any]:
        """
        Handles the MQTT data received from the MQTT broker:
        - parses the event
        - updates node or gateway information
        - returns the event type and data (if any)
        """

        if len(data) < 1:
            return False, EdgeEvent.UNKNOWN, None

        self.last_received_mqtt_data_ts = datetime.now()

        try:
            event_type = EdgeEvent(data[0])
        except ValueError:
            return False, EdgeEvent.UNKNOWN, None

        try:
            if event_type == EdgeEvent.NODE_JOINED:
                node_info = NodeInfoCloud().from_bytes(data[1:])
                if node := self.add_node(node_info.address, node_info.gateway_address):
                    return True, EdgeEvent.NODE_JOINED, node_info

            elif event_type == EdgeEvent.NODE_LEFT:
                node_info = NodeInfoCloud().from_bytes(data[1:])
                if node := self.remove_node(node_info.address, node_info.gateway_address):
                    return True, EdgeEvent.NODE_LEFT, node_info

            elif event_type == EdgeEvent.NODE_KEEP_ALIVE:
                node_info = NodeInfoCloud().from_bytes(data[1:])
                gateway = self.gateways.get(node_info.gateway_address)
                if gateway:
                    gateway.update_node_liveness(node_info.address)
                    return True, EdgeEvent.NODE_KEEP_ALIVE, node_info

            elif event_type == EdgeEvent.GATEWAY_INFO:
                gateway_info = GatewayInfo().from_bytes(data[1:])
                gateway = self.gateways.get(gateway_info.address)
                if not gateway:
                    # we are learning about a new gateway, so instantiate it and add it to the list
                    gateway = MariGateway(info=gateway_info)
                    self.gateways[gateway.info.address] = gateway
                else:
                    gateway.set_info(gateway_info)
                return True, EdgeEvent.GATEWAY_INFO, gateway_info

            elif event_type == EdgeEvent.NODE_DATA:
                frame = Frame().from_bytes(data[1:])

                gateway_address = frame.header.destination
                node_address = frame.header.source
                gateway = self.gateways.get(gateway_address)
                if not gateway:
                    return False, EdgeEvent.UNKNOWN, None
                node = gateway.get_node(node_address)
                if not gateway or not node:
                    return False, EdgeEvent.UNKNOWN, None

                gateway.update_node_liveness(node_address)
                gateway.register_received_frame(frame)

                # handle metrics probe packets
                if frame.is_test_packet:
                    payload = self.metrics_tester.handle_response_cloud(frame, gateway, node)
                    if payload:
                        frame.payload = payload.to_bytes()

                return True, EdgeEvent.NODE_DATA, frame

        except Exception as e:
            print(f"Error handling MQTT data: {e}")

        # fallback result in case of error
        return False, EdgeEvent.UNKNOWN, None

    def on_mqtt_data_received(self, data: bytes):
        res, event_type, event_data = self.handle_mqtt_data(data)
        if res:
            if self.logger and event_type in [EdgeEvent.NODE_JOINED, EdgeEvent.NODE_LEFT]:
                # TODO: update the logging system to also support GATEWAY_INFO events from multiple gateways
                self.logger.log_event(
                    event_data.gateway_address, event_data.address, event_type.name
                )
            self.cb_application(event_type, event_data)



import time
import sys

from marilib.mari_protocol import MARI_NET_ID_DEFAULT
from marilib.marilib_cloud import MarilibCloud
from marilib.communication_adapter import MQTTAdapter
from marilib.model import EdgeEvent


def on_event(event, event_data):
    """An event handler for the application."""
    if event == EdgeEvent.GATEWAY_INFO:
        return
    print(".", end="", flush=True)


def main():
    mari_cloud = MarilibCloud(
        on_event,
        mqtt_interface=MQTTAdapter("localhost", 1883, is_edge=False),
        network_id=int(sys.argv[1], 16) if len(sys.argv) > 1 else MARI_NET_ID_DEFAULT,
    )

    while True:
        mari_cloud.update()
        for node in mari_cloud.nodes:
            # print(f"Sending frame to {node.address:016X}")
            mari_cloud.send_frame(dst=node.address, payload=b"NORMAL_APP_DATA")
            print(",", end="", flush=True)
        statistics = [
            (f"{node.address:016X}", node.stats.received_rssi_dbm()) for node in mari_cloud.nodes
        ]
        print(f"Network statistics: {statistics}")
        time.sleep(1)


if __name__ == "__main__":
    main()



from marilib.mari_protocol import Frame, Header


def test_header_size():
    assert Header().size == 20


def test_header_from_bytes():
    header = Header().from_bytes(
        bytes.fromhex("0210170059291ba8fdcecef531eb7f2526ef0399f0f0f0f0f0")[0:20]
    )
    assert header.version == 2
    assert header.type_ == 16
    assert header.network_id == 23
    assert header.destination == int.from_bytes(
        bytes.fromhex("59291ba8fdcecef5"), byteorder="little"
    )
    assert header.source == int.from_bytes(bytes.fromhex("31eb7f2526ef0399"), byteorder="little")


def test_frame_from_bytes():
    frame = Frame().from_bytes(
        bytes.fromhex("0210170059291ba8fdcecef531eb7f2526ef0399dcf0f0f0f0f0")
    )
    assert frame.header.version == 2
    assert frame.header.type_ == 16
    assert frame.header.network_id == 23
    assert frame.header.destination == int.from_bytes(
        bytes.fromhex("59291ba8fdcecef5"), byteorder="little"
    )
    assert frame.header.source == int.from_bytes(
        bytes.fromhex("31eb7f2526ef0399"), byteorder="little"
    )
    assert frame.stats.rssi_dbm == -35
    assert frame.payload == bytes.fromhex("f0f0f0f0f0")
